import math
import re
import typing as t

import pytest

import rio
import rio.utils


def _assert_stops_match(
    stops_should: t.Iterable[tuple[rio.Color, float]],
    stops_are: t.Iterable[tuple[rio.Color, float]],
) -> None:
    """
    Given two lists of stops (all with positions), verify that they are
    identical.
    """
    stops_should = list(stops_should)
    stops_are = list(stops_are)

    assert len(stops_are) == len(stops_should)

    for stop_should, stop_are in zip(stops_should, stops_are):
        assert isinstance(stop_should, tuple)
        assert isinstance(stop_are, tuple)

        assert len(stop_should) == 2
        assert len(stop_are) == 2

        color_should, position_should = stop_should
        color_are, position_are = stop_are

        assert color_should.hexa == color_are.hexa
        assert math.isclose(position_should, position_are)


@pytest.mark.parametrize(
    "input_type,output_type",
    [
        # Simple cases
        ("pdf", "pdf"),
        (".Pdf", "pdf"),
        (".PDF", "pdf"),
        ("*pdf", "pdf"),
        ("*.Pdf", "pdf"),
        ("*.PDF", "pdf"),
        ("application/pdf", "pdf"),
        # Make sure the results are standardized
        (".jpg", "jpg"),
        (".jpeg", "jpg"),
        ("image/jpeg", "jpg"),
        # Invalid MIME types
        ("not/a/real/type", "type"),
        ("////Type", "type"),
    ],
)
def test_standardize_file_types(
    input_type: str,
    output_type: str,
) -> None:
    cleaned_type = rio.utils.normalize_file_extension(input_type)
    assert cleaned_type == output_type


@pytest.mark.parametrize(
    "unsorted_sequence,keys,sorted_sequence_should",
    [
        (
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
        ),
        (
            [4, 3, 2, 1, 0],
            [4, 3, 2, 1, 0],
            [0, 1, 2, 3, 4],
        ),
        (
            [4, 3, 2, 1, 0],
            [4, None, None, None, 0],
            [0, 3, 2, 1, 4],
        ),
        (
            [0, 1, 2, 3],
            [0, 1, 1, 3],
            [0, 1, 2, 3],
        ),
        (
            [3, 2, 1, 0],
            [3, 1, 1, 0],
            [0, 2, 1, 3],
        ),
        (
            [0, 1, 2, 3],
            [0, None, None, None],
            [0, 1, 2, 3],
        ),
        (
            [0, 1, 2, 3],
            [9, None, None, None],
            [1, 2, 3, 0],
        ),
    ],
)
def test_soft_sort(
    unsorted_sequence: list[int],
    keys: list[int],
    sorted_sequence_should: list[int],
) -> None:
    # Sort the sequence
    sorted_sequence_is = unsorted_sequence.copy()
    rio.utils.soft_sort(
        sorted_sequence_is,
        key=lambda x: keys[unsorted_sequence.index(x)],
    )

    # Verify the result
    assert sorted_sequence_is == sorted_sequence_should


@pytest.mark.parametrize(
    "stops_in, stops_should",
    [
        # No explicit positions
        (
            (rio.Color.BLACK,),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.BLACK, 1.0),
            ),
        ),
        (
            (
                rio.Color.BLACK,
                rio.Color.WHITE,
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        (
            (
                rio.Color.BLACK,
                rio.Color.BLUE,
                rio.Color.WHITE,
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.BLUE, 0.5),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        # Containing explicit positions
        (
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.WHITE, 1.0),
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        (
            (
                (rio.Color.BLACK, 0.0),
                rio.Color.WHITE,
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        (
            (
                (rio.Color.BLACK, 0.0),
                rio.Color.BLUE,
                rio.Color.WHITE,
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.BLUE, 0.5),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        (
            (
                (rio.Color.BLACK, 0.0),
                rio.Color.RED,
                rio.Color.GREEN,
                rio.Color.BLUE,
                (rio.Color.CYAN, 0.8),
                rio.Color.MAGENTA,
                (rio.Color.WHITE, 1.0),
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.RED, 0.2),
                (rio.Color.GREEN, 0.4),
                (rio.Color.BLUE, 0.6),
                (rio.Color.CYAN, 0.8),
                (rio.Color.MAGENTA, 0.9),
                (rio.Color.WHITE, 1.0),
            ),
        ),
        # Duplicate positions
        (
            (
                rio.Color.BLACK,
                (rio.Color.RED, 0.5),
                (rio.Color.GREEN, 0.5),
                rio.Color.WHITE,
            ),
            (
                (rio.Color.BLACK, 0.0),
                (rio.Color.RED, 0.5),
                (rio.Color.GREEN, 0.5),
                (rio.Color.WHITE, 1.0),
            ),
        ),
    ],
)
def test_interpolation_is_correct(
    stops_in: tuple[rio.Color, ...],
    stops_should: tuple[tuple[rio.Color, float], ...],
) -> None:
    """
    Test that the interpolation function correctly assigns positions to stops.
    """
    stops_are = rio.utils.verify_and_interpolate_gradient_stops(stops_in)
    _assert_stops_match(stops_should, stops_are)


@pytest.mark.parametrize(
    "stops_in, error_msg",
    [
        # Empty stops
        (
            tuple(),
            "Gradients must have at least one stop",
        ),
        # Position out of range
        (
            (
                (rio.Color.BLACK, -0.1),
                (rio.Color.WHITE, 1.0),
            ),
            "Gradient stop positions must be in range [0, 1], but stop 0 is at position -0.1",
        ),
        (
            (
                (rio.Color.BLACK, 0.1),
                (rio.Color.WHITE, 1.1),
            ),
            "Gradient stop positions must be in range [0, 1], but stop 1 is at position 1.1",
        ),
        # Positions not ascending
        (
            (
                (rio.Color.BLACK, 0.8),
                (rio.Color.WHITE, 0.2),
            ),
            "Gradient stops must be in ascending order, but stop 0 is at position 0.8 while stop 1 is at position 0.2",
        ),
        (
            (
                rio.Color.RED,
                (rio.Color.BLACK, 0.8),
                rio.Color.BLUE,
                (rio.Color.WHITE, 0.2),
            ),
            "Gradient stops must be in ascending order, but stop 1 is at position 0.8 while stop 3 is at position 0.2",
        ),
    ],
)
def test_invalid_stops_raise_error(
    stops_in: tuple[rio.Color | tuple[rio.Color, float], ...],
    error_msg: str,
) -> None:
    """
    Test that invalid gradient stop configurations raise appropriate errors.
    """
    with pytest.raises(ValueError, match=re.escape(error_msg)):
        rio.utils.verify_and_interpolate_gradient_stops(stops_in)



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
NAV_BUTTON_DATA_MOBILE: list[tuple[str, str]] = [
    ("Home", ""),
    ("Pricing", "pricing"),
    ("Blog", "blog"),
]


class HamburgerButton(rio.Component):
    """
    A toggleable hamburger button component.

    This component displays a hamburger menu icon when the menu is closed and a
    close icon when the menu is open. It allows toggling the `is_open` state
    through a pointer (click/touch) event.


    ## Attributes:

    `is_open`: Indicates whether the menu is currently open or closed.
    """

    is_open: bool = False

    def _on_toggle_open(self, _: rio.PointerEvent) -> None:
        """
        Handles the toggle action when the button is pressed.

        This method toggles the `is_open` state between `True` and `False`.
        """
        self.is_open = not self.is_open

    def build(self) -> rio.Component:
        """
        Builds the hamburger button component.

        The button dynamically changes its icon and appearance based on the
        `is_open` state.
        """
        # Determine the icon and fill color based on the open state
        if self.is_open:
            # Display a close icon when the menu is open
            icon = "material/close"
            fill = self.session.theme.text_style.fill
        else:
            # Display a hamburger menu icon when closed
            icon = "material/menu"
            fill = self.session.theme.text_style.fill

        # Ensure the fill is a valid rio.Color instance
        assert isinstance(fill, rio.Color)

        return rio.PointerEventListener(
            rio.Rectangle(
                content=rio.Icon(
                    icon=icon,
                    fill=fill,
                    min_width=2.2,
                    min_height=2.2,
                ),
                fill=rio.Color.TRANSPARENT,
                # Show pointer cursor on hover
                cursor=rio.CursorStyle.POINTER,
            ),
            on_press=self._on_toggle_open,
        )


class Navbar(rio.Component):
    """
    A responsive and fixed-position navigation bar with separate layouts for
    desktop and mobile devices.


    ## Attributes:

    `is_open`: Whether the mobile navigation menu is open.
    """

    is_open: bool = False

    # Make sure the navbar will be rebuilt when the app navigates to a different
    # page. While Rio automatically detects state changes and rebuilds
    # components as needed, navigating to other pages is not considered a state
    # change, since it's not stored in the component.
    #
    # Instead, we can use Rio's `on_page_change` event to trigger a rebuild of
    # the navbar when the page changes.
    @rio.event.on_page_change
    def on_page_change(self) -> None:
        """
        Trigger a rebuild of the navbar when the page changes.
        """
        # Rio comes with a function specifically for this. Whenever Rio is
        # unable to detect a change automatically, use this function to force a
        # refresh.
        self.force_refresh()

    def _get_active_url_fragment(self) -> str | None:
        """
        Get the active URL fragment to determine the currently selected page.
        """

        # Prepare the URL
        segments = self.session.active_page_url.parts

        # Special case: Not deep enough
        if len(segments) <= 1:
            return None

        # Special case: Home page
        if segments[1] == "":
            return ""

        # Regular case: use the url segment of the active page
        return segments[1]

    def _on_switcherbar_value_change(
        self, event: rio.SwitcherBarChangeEvent
    ) -> None:
        """
        Handles changes to the SwitcherBar value.

        Navigates to the corresponding page and closes the mobile menu.
        """
        # Prepare the URL
        if event.value is None:
            target_segment = self._get_active_url_fragment()
        else:
            target_segment = event.value

        # Navigate to the target page
        self.session.navigate_to(f"/{target_segment}")
        # Close the mobile menu
        self.is_open = False

    def _desktop_build(self) -> rio.Component:
        """
        Build the desktop layout.
        """
        # Which page is currently active? This will be used to highlight the
        # correct navigation button.
        #
        # `active_page_instances` contains `rio.ComponentPage` instances that
        # that are created during app creation. Since multiple pages can be
        # active at a time (e.g. /foo/bar/baz), this is a list rather than just
        # a single page.
        active_page = self.session.active_page_instances[0]
        active_page_url_segment = active_page.url_segment

        # Determine styling based on theme (light or dark)
        if self.session.theme.is_light_theme:
            fine_border_line_fill_color = self.session.theme.primary_color
            frosted_glass_fill_color = (
                self.session.theme.background_color.replace(opacity=0.7)
            )

        else:
            fine_border_line_fill_color = self.session.theme.neutral_color
            frosted_glass_fill_color = (
                self.session.theme.background_color.replace(opacity=0.7)
            )

        # The navbar should appear above all other components. This is easily
        # done by using a `rio.Overlay` component.
        return rio.Overlay(
            rio.Column(
                # The navbar consists of a rectangle with a row of buttons inside.
                rio.Rectangle(
                    content=rio.Row(
                        rio.Spacer(),
                        rio.Row(
                            # By sticking buttons into a `rio.Link`, we can easily
                            # make the buttons navigate to other pages, without
                            # having to write an event handler. Notice how there is
                            # no Python function called when the button is clicked.
                            rio.Link(
                                rio.IconButton(
                                    "rio/logo",
                                    style="plain-text",
                                    min_size=2.5,
                                ),
                                "/",
                            ),
                            rio.Spacer(),
                            # Same game different button
                            rio.Link(
                                rio.Button(
                                    "Pricing",
                                    style=(
                                        "colored-text"
                                        if active_page_url_segment == "pricing"
                                        else "plain-text"
                                    ),
                                ),
                                "/pricing",
                            ),
                            # Same game, different button
                            rio.Link(
                                rio.Button(
                                    "Blog",
                                    style=(
                                        "colored-text"
                                        if active_page_url_segment == "blog"
                                        else "plain-text"
                                    ),
                                ),
                                "/blog",
                            ),
                            # Same game, different button
                            comps.OutlinedNeutralButton(text="Sign in"),
                            # Same game, different button
                            rio.Link(
                                content=rio.Button(
                                    rio.Row(
                                        rio.Text(
                                            "Sign up",
                                            style=rio.TextStyle(
                                                fill=self.session.theme.background_color
                                            ),
                                        ),
                                        rio.Icon("material/arrow_forward"),
                                        align_x=0.5,
                                        spacing=1,
                                        margin_x=1,
                                    ),
                                    color=rio.Color.from_hex("ffffff"),
                                ),
                                target_url="/",  # TODO: Change this to the correct URL
                            ),
                            spacing=1,
                            margin_y=1,
                            min_width=80,
                            align_x=0.5,
                        ),
                        rio.Spacer(),
                    ),
                    # Set the fill of the rectangle to the neutral color of the theme and
                    # add a FrostedGlassFill
                    fill=rio.FrostedGlassFill(
                        frosted_glass_fill_color,
                        blur_size=0.6,
                    ),
                    min_height=4.5,
                    align_y=0,
                ),
                # Add a thin line to separate the navbar from the rest of the content
                # This line will be placed below the rectangle
                rio.Rectangle(
                    fill=fine_border_line_fill_color,
                    min_height=0.1,
                    align_y=0,
                ),
                spacing=0,
                align_y=0,
            ),
        )

    def _mobile_build(self) -> rio.Component:
        """
        Build the mobile layout.
        """
        active_url_fragment = self.session.active_page_url.name

        # Determine styling based on theme (light or dark)
        if self.session.theme.is_light_theme:
            fine_border_line_fill_color = self.session.theme.primary_color
            frosted_glass_fill_color = (
                self.session.theme.background_color.replace(opacity=0.7)
            )

        else:
            fine_border_line_fill_color = self.session.theme.neutral_color
            frosted_glass_fill_color = (
                self.session.theme.background_color.replace(opacity=0.7)
            )

        # Add the mobile navigation menu when open
        if self.is_open:
            # Add the local navigation buttons
            primary_column = rio.Rectangle(
                content=rio.SwitcherBar(
                    names=[option[0] for option in NAV_BUTTON_DATA_MOBILE],
                    values=[option[1] for option in NAV_BUTTON_DATA_MOBILE],
                    selected_value=active_url_fragment,
                    allow_none=True,
                    color="primary",
                    orientation="vertical",
                    on_change=self._on_switcherbar_value_change,
                    margin=0.5,
                    align_y=0,
                ),
                fill=self.session.theme.neutral_color,
                corner_radius=(
                    0,
                    0,
                    self.session.theme.corner_radius_large,
                    self.session.theme.corner_radius_large,
                ),
                shadow_radius=1,
                margin_bottom=1,
            )

        else:
            primary_column = None

        # The navbar should appear above all other components. This is easily
        # done by using a `rio.Overlay` component.
        return rio.Overlay(
            rio.Column(
                rio.Rectangle(
                    content=rio.Row(
                        rio.Link(
                            rio.IconButton(
                                "rio/logo",
                                style="plain-text",
                                min_size=2.5,
                            ),
                            "/",
                        ),
                        rio.Spacer(),
                        comps.OutlinedNeutralButton(
                            text="Sign in",
                            align_y=0.5,
                        ),
                        HamburgerButton(is_open=self.bind().is_open),
                        spacing=1.1,
                        margin_x=0.6,
                        margin_left=1.2,
                        margin_right=1.2,
                    ),
                    # Set the fill of the rectangle to the neutral color of the theme and
                    # add a FrostedGlassFill
                    fill=rio.FrostedGlassFill(
                        frosted_glass_fill_color,
                        blur_size=0.6,
                    ),
                    align_y=0,
                    min_height=4,  # Fixed height of the navbar
                ),
                # Add a thin line to separate the navbar from the rest of the content
                # This line will be placed below the rectangle
                rio.Rectangle(
                    fill=fine_border_line_fill_color,
                    min_height=0.1,
                    align_y=0,
                ),
                rio.Switcher(content=primary_column),
                align_y=0,
            ),
        )

    def build(self) -> rio.Component:
        """
        Build and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
@rio.page(
    name="Home",
    url_segment="",
)
class LandingPage(rio.Component):
    """
    The landing page of the website.

    This page features the main content of the website, including the hero
    section, services of our fictional company, cards, testimonials, and a 'Get
    Started' call-to-action section.
    """

    def _desktop_build(self) -> rio.Component:
        """
        Build the layout optimized for desktop devices.
        """
        return rio.Column(
            comps.HeroSection(),
            comps.ServicePower(),
            comps.ServiceSpeed(),
            comps.CardSection(),
            comps.Testimonials(),
            comps.GetStarted(),
            spacing=14,  # Space between components for better readability on desktop
            align_x=0.5,  # Center horizontally
            align_y=0,  # Align to the top vertically
            min_width=80,  # Fixed width to ensure proper layout on larger screens
            margin_bottom=10,  # Space below the column
        )

    def _mobile_build(self) -> rio.Component:
        """
        Build the layout optimized for mobile devices.
        """
        return rio.Column(
            comps.HeroSection(),
            comps.ServicePower(),
            comps.ServiceSpeed(),
            comps.CardSection(),
            comps.Testimonials(),
            comps.GetStarted(),
            spacing=7,  # Reduced space to accommodate smaller screens
            margin=1,  # Minimal margin for compact layout
        )

    def build(self) -> rio.Component:
        """
        Construct and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



from __future__ import annotations

import abc
import dataclasses
import functools
import logging
import typing as t
import warnings
from pathlib import Path

import imy.docstrings
import introspection.typing
import path_imports
from introspection import convert_case

import rio.components.error_placeholder

from . import deprecations, url_pattern, utils
from .errors import NavigationFailed

__all__ = [
    "Redirect",
    "ComponentPage",
    "page",
    "GuardEvent",
    "QueryParameter",
]


DEFAULT_ICON = "rio/logo:color"


class QUERY_PARAMETER:
    pass


T = t.TypeVar("T")
QueryParameter = t.Annotated[T, QUERY_PARAMETER]


def _verify_url_and_parse_into_pattern(
    url_segment: str,
) -> url_pattern.UrlPattern:
    """
    Runs some checks on the given URL segment. If they pass, the URL segment
    is parsed into a URL pattern object, which is returned.
    """
    # In Rio, URLs are case insensitive. An easy way to enforce this, and
    # also prevent casing issues in the user code is to make sure the page's
    # URL fragment is lowercase.
    if url_segment != url_segment.lower():
        raise ValueError(
            f"Page URL segments should be lowercase, but `{url_segment}` is not"
        )

    # Parse the url segment into an URL pattern object
    return url_pattern.UrlPattern(url_segment)


@t.final
@dataclasses.dataclass(frozen=True)
class Redirect:
    """
    Redirects the user to a different page.

    Redirects can be added to the app in place of "real" pages. These are
    useful for cases where you want to add valid links to the app, but don't
    want to display a page at that URL. They redirect users to another page
    the instant they would be opened.

    Redirects are passed directly to the app during construction, like so:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my app!"),
            rio.PageView(grow_y=True),
        ),
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.ComponentPage(
                name="Subpage",
                url_segment="subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
            rio.Redirect(
                url_segment="old-page",
                target="/subpage",
            ),
        ],
    )

    app.run_in_browser()
    ```

    ## Attributes

    `url_segment`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `target`: The URL to which the user should be redirected.
    """

    url_segment: str
    target: str | rio.URL

    # A pre-parsed URL pattern object, used to verify whether a URL matches
    # this page, as well as extracting path parameters
    _url_pattern: url_pattern.UrlPattern = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        vars(self).update(
            # Verify and parse the URL
            _url_pattern=_verify_url_and_parse_into_pattern(self.url_segment)
        )


@t.final
@deprecations.parameter_renamed(
    since="0.10",
    old_name="page_url",
    new_name="url_segment",
)
@dataclasses.dataclass(frozen=True)
class ComponentPage:
    """
    A routable page in a Rio app.

    Rio apps can consist of many pages. You might have a welcome page, a
    settings page, a login, and so on. `ComponentPage` components contain all
    information needed to display those pages, as well as to navigate between
    them.

    This is not just specific to websites. Apps might, for example, have
    a settings page, a profile page, a help page, and so on.

    Pages are passed directly to the app during construction, like so:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my app!"),
            rio.PageView(grow_y=True),
        ),
        pages=[
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.ComponentPage(
                name="Subpage",
                url_segment="subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
        ],
    )

    app.run_in_browser()
    ```

    This will display "This is the home page" when navigating to the root URL,
    but "This is a subpage" when navigating to "/subpage". Note that on both
    pages the text "Welcome to my page!" is displayed above the page content.
    That's because it's not part of the `PageView`.

    For additional details, please refer to the how-to guide:
    `https://rio.dev/docs/howto/multiple-pages`.

    ## Attributes

    `name`: A human-readable name for the page. While the page itself doesn't
        use this value directly, it serves as important information for
        debugging, as well as other components such as navigation bars.

    `url_segment`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `build`: A callback that is called when this page is displayed. It should
        return a Rio component.

    `icon`: The name of an icon to associate with the page. While the page
        itself doesn't use this value directly, it serves as additional
        information for other components such as navigation bars.

    `children`: A list of child pages. These pages will be displayed when
        navigating to a sub-URL of this page. For example, if this page's
        `url_segment` is "page1", and it has a child page with `url_segment`
        "page2", then the child page will be displayed at
        "https://yourapp.com/page1/page2".

    `meta_tags`: A dictionary of meta tags to include in the page's HTML. These
        are used by search engines and social media sites to display
        information about your page.

    `guard`: A callback that is called before this page is displayed. It
        can prevent users from accessing pages which they are not allowed to
        see. For example, you may want to redirect users to your login page
        if they are trying to access their profile page without being
        logged in.

        The callback should return `None` if the user is allowed to access
        the page, or a string or `rio.URL` if the user should be redirected
        to a different page.
    """

    name: str
    url_segment: str
    build: t.Callable[..., rio.Component]
    _: dataclasses.KW_ONLY
    icon: str = DEFAULT_ICON
    children: t.Sequence[ComponentPage | Redirect] = dataclasses.field(
        default_factory=list
    )
    guard: t.Callable[[rio.GuardEvent], None | rio.URL | str] | None = None
    meta_tags: dict[str, str] = dataclasses.field(default_factory=dict)

    # This is used to allow users to order pages when using the `rio.page`
    # decorator. It's not public, but simply a convenient place to store this.
    _page_order_: int | None = dataclasses.field(default=None, init=False)

    # A pre-parsed URL pattern object, used to verify whether a URL matches
    # this page, as well as extracting path parameters
    _url_pattern: url_pattern.UrlPattern = dataclasses.field(init=False)

    # The names of the query parameters that are passed to the `build` function
    _url_parameter_parsers: t.Mapping[str, UrlParameterParser] = (
        dataclasses.field(init=False)
    )

    def __post_init__(self) -> None:
        # Verify and parse the URL
        vars(self)["_url_pattern"] = _verify_url_and_parse_into_pattern(
            self.url_segment
        )

        # Verify the `build` function and prepare the parsers for the URL
        # parameters
        try:
            vars(self)["_url_parameter_parsers"] = (
                self._verify_build_function_and_get_parameter_parsers()
            )
        except TypeError as error:
            raise TypeError(
                f"{self.build} is not a valid `build` function: {error}"
            ) from None

    def _verify_build_function_and_get_parameter_parsers(
        self,
    ) -> t.Mapping[str, UrlParameterParser]:
        # Figure out which query parameters we need to pass to the `build`
        # function and how to parse them
        signature = introspection.signature(self.build)
        url_parameter_parsers = {}

        for param_name, parameter in signature.parameters.items():
            # Is this a path parameter, a query parameter, or neither?
            try:
                type_info = introspection.typing.TypeInfo(
                    parameter.annotation,
                    forward_ref_context=parameter.forward_ref_context,
                )
            except introspection.errors.CannotResolveForwardref:
                # If it seems likely that this was supposed to be a query
                # parameter, emit a warning. Otherwise, silently ignore it.
                if isinstance(parameter.annotation, t.ForwardRef):
                    annotation_str = parameter.annotation.__forward_arg__
                else:
                    annotation_str = str(parameter.annotation)

                annotation_str = annotation_str.lower()

                if "query" in annotation_str or "param" in annotation_str:
                    warnings.warn(
                        f"The type annotation of the {param_name!r} parameter"
                        f" of the {self.build.__name__!r} `build` function"
                        f" cannot be resolved. If this was intended to be a"
                        f" query parameter, make sure the annotation is valid"
                        f" at runtime."
                    )

                continue

            if param_name not in self._url_pattern.path_parameter_names:
                if QUERY_PARAMETER not in type_info.annotations:
                    continue

                # Query parameters must have default values
                if not parameter.is_optional:
                    raise TypeError(
                        f"The query parameter {param_name!r} doesn't have a default value"
                    )

            if parameter.kind not in (
                introspection.Parameter.POSITIONAL_OR_KEYWORD,
                introspection.Parameter.KEYWORD_ONLY,
            ):
                raise TypeError(
                    f"The {param_name!r} parameter isn't a keyword parameter"
                )

            if parameter.annotation is introspection.Parameter.empty:
                raise TypeError(
                    f"The {param_name!r} parameter doesn't have a type annotation"
                )

            url_parameter_parsers[param_name] = _get_parser_for_annotation(
                type_info
            )

        # Make sure no parameters that were defined in the URL pattern are
        # missing
        for param_name in self._url_pattern.path_parameter_names:
            if param_name not in signature.parameters:
                raise TypeError(
                    f"{self.build} is not a valid `build` function: The URL pattern defines a parameter {param_name!r}, but the function doesn't have such a parameter"
                )

        return url_parameter_parsers

    def _safe_build_with_url_parameters(
        self,
        path_params: t.Mapping[str, object],
        raw_query_params: t.Mapping[str, str],
    ) -> rio.Component:
        parsed_query_params = self._parse_query_parameters(raw_query_params)
        return utils.safe_build(
            self.build, **path_params, **parsed_query_params
        )

    def _parse_query_parameters(
        self, query_params: t.Mapping[str, str]
    ) -> t.Mapping[str, object]:
        parsed_params = dict[str, object]()

        for name, raw_value in query_params.items():
            try:
                parser = self._url_parameter_parsers[name]
            except KeyError:
                continue

            try:
                parsed_params[name] = parser.parse(raw_value)
            except ValueError:
                pass

        return parsed_params


@deprecations.deprecated(since="0.10", replacement=ComponentPage)
@introspection.set_signature(ComponentPage)
def Page(*args, **kwargs):
    return ComponentPage(*args, **kwargs)


class UrlParameterParser(abc.ABC):
    @abc.abstractmethod
    def parse(self, value: str) -> object:
        raise NotImplementedError

    @abc.abstractmethod
    def list_valid_values(self) -> t.Sequence[str]:
        """
        List all values that can be parsed by this parser, if there is a
        reasonable amount of them. (Don't list all floats, for example.) If
        there are too many, throw a `ValueError`.
        """
        raise NotImplementedError


class FuncParser(UrlParameterParser):
    """
    A simple parser that simply calls a function to parse the value.
    `list_valid_values()` always throws a `ValueError`.
    """

    def __init__(self, parse_function: t.Callable[[str], object]):
        self._parse = parse_function

    def parse(self, value: str) -> object:
        return self._parse(value)

    def list_valid_values(self):
        raise ValueError


class BooleanParser(UrlParameterParser):
    def parse(self, value: str) -> bool:
        value = value.lower()

        if value in ("true", "t", "yes", "y", "1"):
            return True

        if value in ("false", "f", "no", "n", "0"):
            return False

        raise ValueError(f"{value!r} cannot be parsed as a boolean")

    def list_valid_values(self):
        return ("true", "false")


class LiteralParser(UrlParameterParser):
    def __init__(self, values: tuple[object, ...]):
        self.values = values

        types = {type(value) for value in values}
        if len(types) > 1:
            raise TypeError(
                "`Literal`s with values of different types aren't supported"
            )

        parameter_type = types.pop()
        try:
            self._sub_parser = _get_parser_for_annotation(
                introspection.typing.TypeInfo(parameter_type)
            )
        except TypeError:
            raise TypeError(
                f"`Literal`s of type {parameter_type.__name__} aren't supported"
            ) from None

    def parse(self, value: str):
        parsed_value = self._sub_parser.parse(value)

        if parsed_value in self.values:
            return parsed_value

        raise ValueError(f"{value!r} doesn't match any of the allowed literals")

    def list_valid_values(self) -> t.Sequence[str]:
        return [str(value) for value in self.values]


def _get_parser_for_annotation(
    annotation: introspection.typing.TypeInfo,
) -> UrlParameterParser:
    TYPE_TO_PARSER_FUNC: t.Mapping[
        introspection.types.TypeAnnotation, t.Callable[[str], object]
    ] = {
        str: str,
        int: int,
        float: float,
    }

    try:
        parser_func = TYPE_TO_PARSER_FUNC[annotation.type]
    except KeyError:
        pass
    else:
        return FuncParser(parser_func)

    if annotation.type == bool:
        return BooleanParser()

    if annotation.type == t.Literal:
        assert annotation.arguments
        return LiteralParser(annotation.arguments)

    # As a special case, `None` is allowed in type annotations because it's a
    # popular default value, but we won't actually parse it. It can *only* be
    # used as a default value.
    if annotation.type == t.Optional:
        subtype: introspection.types.TypeAnnotation = annotation.arguments[0]  # type: ignore
        return _get_parser_for_annotation(
            introspection.typing.TypeInfo(
                subtype, forward_ref_context=annotation.forward_ref_context
            )
        )

    raise TypeError(
        f"Parameters of type {annotation.type} aren't supported"
    ) from None


def _get_active_page_instances(
    *,
    available_pages: t.Iterable[rio.ComponentPage | rio.Redirect],
    remaining_path: str,
) -> list[
    tuple[
        rio.ComponentPage | rio.Redirect,
        dict[str, object],
    ]
]:
    """
    Given a list of available pages, and a URL, return the list of pages that
    would be active if navigating to that URL. Each result entry contains the

    - Page (ComponentPage / Redirect) that would be active
    - The path arguments passed to that page

    The path is a string rather than URL, so matching can be done efficiently
    and the function can recurse on it. The path string must not start with a
    slash.
    """
    assert not remaining_path.startswith("/"), remaining_path

    # Get the first matching page
    for page in available_pages:
        did_match, raw_path_arguments, remainder = page._url_pattern.match(
            remaining_path
        )

        if not did_match:
            continue

        # Try parsing the path arguments
        if isinstance(page, rio.ComponentPage):
            try:
                path_arguments = {
                    name: page._url_parameter_parsers[name].parse(value)
                    for name, value in raw_path_arguments.items()
                }
            except ValueError:
                continue

        else:
            path_arguments = raw_path_arguments

        break

    # No matching page found
    else:
        return []

    # Remember this page
    path_arguments = t.cast(dict[str, object], path_arguments)
    active_pages = [
        (page, path_arguments),
    ]

    # Recurse into the children
    if isinstance(page, rio.ComponentPage):
        active_pages += _get_active_page_instances(
            available_pages=page.children,
            remaining_path=remainder,
        )

    return active_pages


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass(frozen=True)
class GuardEvent:
    """
    Holds information regarding a guard event.

    This is a simple dataclass that stores useful information for the guard
    event. They can prevent users from accessing pages which they are not
    allowed to see.

    ## Attributes

    `session`: The current session.

    `active_pages`: All pages that will be active if the navigation succeeds.
    """

    # The current session
    session: rio.Session

    # The pages that would be activated by this navigation
    #
    # This is an `Sequence` rather than `list`, because the same event instance
    # is reused for multiple event handlers. This allows to assign a tuple, thus
    # preventing modifications.
    active_pages: t.Sequence[ComponentPage | Redirect]


def check_page_guards(
    sess: rio.Session,
    target_url_absolute: rio.URL,
) -> tuple[
    tuple[
        tuple[ComponentPage, dict[str, object]],
        ...,
    ]
    | None,
    rio.URL,
]:
    """
    Check whether navigation to the given target URL is possible.

    This finds the pages that would be activated by this navigation and runs
    their guards. If the guards effect a redirect, it instead attempts to
    navigate to the redirect target, and so on.

    Raises `NavigationFailed` if navigation to the target URL is not possible
    because of an error, such as an exception in a guard.

    If the URL points to a page which doesn't exist that is not considered an
    error. The result will still be valid. That is because navigation is
    possible, it's just that some `PageView`s will display a 404 page.

    If attempting to navigate to a URL that isn't path of the app, i.e. it's not
    a child of the app's configured base URL, `None` is returned instead of
    active pages.

    This function does not perform any actual navigation. It simply checks
    whether navigation to the target page is possible.

    The result is a tuple of two elements:

    - A tuple of tuples, where each inner tuple contains a page that would be
      activated by the navigation, and the path arguments that would be passed
      to that page during building.

    - The URL that the navigation would actually lead to. This is the URL that
      the user would see in their browser's address bar after the navigation
      completes.
    """

    assert target_url_absolute.is_absolute(), target_url_absolute

    # Is any guard opposed to this page?
    initial_target_url = target_url_absolute
    visited_redirects = {target_url_absolute}
    past_redirects = [target_url_absolute]

    while True:
        # Strip the base. This also detects navigation to outside of the app
        try:
            target_url_relative = utils.url_relative_to_base(
                sess._base_url, target_url_absolute
            )
        except ValueError:
            return None, target_url_absolute

        # Find all pages which would by activated by this navigation
        active_page_instances_and_path_arguments = tuple(
            _get_active_page_instances(
                available_pages=sess.app.pages,
                remaining_path=target_url_relative.path,
            )
        )

        # Check the guards for each activated page
        active_page_instances = tuple(
            page for page, _ in active_page_instances_and_path_arguments
        )

        redirect = None
        event = GuardEvent(
            session=sess,
            active_pages=active_page_instances,
        )

        for page in active_page_instances:
            if isinstance(page, Redirect):
                redirect = page.target
            else:
                if page.guard is None:
                    continue

                try:
                    redirect = page.guard(event)
                except Exception as err:
                    message = f"Rejecting navigation to `{initial_target_url}` because of an error in a page guard of `{page.url_segment}`: {err}"
                    logging.exception(message)
                    raise NavigationFailed(message)

            if redirect is not None:
                break

        # All guards are satisfied - done
        if redirect is None:
            assert all(
                isinstance(page, rio.ComponentPage)
                for page in active_page_instances
            ), active_page_instances

            return (
                active_page_instances_and_path_arguments,  # type: ignore (there cannot be any Redirects here anymore)
                target_url_absolute,
            )

        # A guard wants to redirect to a different page
        redirect = sess._make_url_absolute(
            redirect, active_page_url_override=target_url_absolute
        )
        assert redirect.is_absolute(), redirect

        # Detect infinite loops and break them
        if redirect in visited_redirects:
            page_strings = [
                str(url_segment) for url_segment in past_redirects + [redirect]
            ]
            page_strings_list = "\n -> ".join(page_strings)

            message = f"Rejecting navigation to `{initial_target_url}` because page guards have created an infinite loop:\n\n    {page_strings_list}"
            logging.warning(message)
            raise NavigationFailed(message)

        # Remember that this page has been visited before
        visited_redirects.add(redirect)
        past_redirects.append(redirect)

        # Rinse and repeat
        target_url_absolute = redirect


BuildFunction = t.Callable[..., "rio.Component"]
C = t.TypeVar("C", bound=BuildFunction)


BUILD_FUNCTION_TO_PAGE = dict[BuildFunction, ComponentPage]()


def page(
    *,
    url_segment: str | None = None,
    name: str | None = None,
    icon: str = DEFAULT_ICON,
    guard: t.Callable[[GuardEvent], None | rio.URL | str] | None = None,
    meta_tags: dict[str, str] | None = None,
    order: int | None = None,
) -> t.Callable[[C], C]:
    """
    This decorator creates a page (complete with URL, icon, etc) that displays
    the decorated component. All parameters are optional, and if omitted,
    sensible defaults will be inferred based on the name of the decorated class.

    In order to create a "root" page, set the `url_segment` to an empty string:

    ```py
    @rio.page(
        url_segment="",
    )
    class HomePage(rio.Component):
        def build(self):
            return rio.Text(
                "Welcome to my website",
                style="heading1",
            )
    ```

    For additional details, please refer to the how-to guide [Multiple
    Pages](https://rio.dev/docs/howto/multiple-pages).


    ## Parameters

    `url_segment`: The URL segment at which this page should be displayed. For
        example, if this is "subpage", then the page will be displayed at
        "https://yourapp.com/subpage". If this is "", then the page will be
        displayed at the root URL.

    `name`: A human-readable name for the page. While the page itself doesn't
        use this value directly, it serves as important information for
        debugging, as well as other components such as navigation bars.

    `icon`: The name of an icon to associate with the page. While the page
        itself doesn't use this value directly, it serves as additional
        information for other components such as navigation bars.

    `meta_tags`: A dictionary of meta tags to include in the page's HTML. These
        are used by search engines and social media sites to display
        information about your page.

    `guard`: A callback that is called before this page is displayed. It
        can prevent users from accessing pages which they are not allowed to
        see. For example, you may want to redirect users to your login page
        if they are trying to access their profile page without being
        logged in.

        The callback should return `None` if the user is allowed to access
        the page, or a string or `rio.URL` if the user should be redirected
        to a different page.

    `order`: An int that controls the order of this page relative to its
        siblings. Similar to the `name`, this is relevant for navigation bars.
    """

    def decorator(build: C) -> C:
        nonlocal name, url_segment

        # Derive a default name
        if name is None:
            name = (
                convert_case(build.__name__, "snake").replace("_", " ").title()
            )

        # Derive a default URL segment
        if url_segment is None:
            url_segment = convert_case(build.__name__, "kebab").lower()

        # Create the result
        page = ComponentPage(
            name=name,
            url_segment=url_segment,
            build=build,
            icon=icon,
            guard=guard,
            meta_tags=meta_tags or {},
        )

        # The component page has a field specifically so this decorator can
        # store the page order. However, since this is a frozen dataclass,
        # contortions are needed
        page.__dict__["_page_order_"] = order

        # Store the result
        BUILD_FUNCTION_TO_PAGE[build] = page

        # Return the original class
        return build

    return decorator


def _page_sort_key(page: rio.ComponentPage) -> tuple:
    """
    Returns a key that can be used to sort pages.
    """
    return (
        # Put the home page first
        not page.url_segment == "",
        # Then sort by name
        page.name,
    )


def auto_detect_pages(
    directory: Path,
    *,
    package: str | None = None,
) -> list[rio.ComponentPage]:
    # Find all pages using the iterator method
    pages = _auto_detect_pages_iter(directory, package=package)

    # Sort them, ignoring any user-specified ordering for now
    pages = sorted(pages, key=_page_sort_key)

    # Now apply the user-specified ordering. This sorting is stable, hence the
    # previous step.
    utils.soft_sort(pages, key=lambda page: page._page_order_)

    # Done
    return pages


def _auto_detect_pages_iter(
    directory: Path,
    *,
    package: str | None = None,
) -> t.Iterable[rio.ComponentPage]:
    try:
        contents = list(directory.iterdir())
    except FileNotFoundError:
        return

    for file_path in contents:
        if not utils.is_python_script(file_path):
            continue

        yield _page_from_python_file(file_path, package)


def _page_from_python_file(
    file_path: Path, package: str | None
) -> ComponentPage:
    module_name = file_path.stem
    if package is not None:
        module_name = package + "." + module_name

    try:
        module = path_imports.import_from_path(
            file_path,
            module_name=module_name,
            force_reimport=False,
        )
    except BaseException as error:
        import traceback

        traceback.print_exc()

        # Can't import the module? Display a warning and a placeholder component
        warnings.warn(
            f"Failed to import file '{file_path}': {type(error)} {error}"
        )
        page = _error_page_from_file_name(
            file_path,
            error_summary=f"Failed to import '{file_path}'",
            error_details=f"{type(error).__name__}: {error}",
        )
    else:
        # Search the module for the callable decorated with `@rio.page`
        pages = list[ComponentPage]()
        for obj in vars(module).values():
            try:
                pages.append(BUILD_FUNCTION_TO_PAGE[obj])
            except (TypeError, KeyError):
                pass

        if pages:
            page = pages[0]

            # More than one page found? Display a warning
            if len(pages) > 1:
                warnings.warn(
                    f"The file {file_path} contains multiple page definitions."
                    f" This is not allowed. Each page must be defined in its"
                    f" own file."
                )
        else:
            # Nothing found? Display a warning and a placeholder component
            warnings.warn(
                f"The file {file_path} doesn't seem to contain a page"
                f" definition. Did you forget to decorate your component/build"
                f" function with `@rio.page(...)`?"
            )
            page = _error_page_from_file_name(
                file_path,
                error_summary=f"No page found in '{file_path}'",
                error_details=f"No component in this file was decorated with `@rio.page(...)`",
            )

    # Add sub-pages, if any
    sub_pages = t.cast(list, page.children)
    sub_pages.clear()  # Avoid duplicate entries if this function is called twice
    sub_pages += auto_detect_pages(
        file_path.with_suffix(""),
        package=module_name,
    )

    return page


def _error_page_from_file_name(
    file_path: Path, error_summary: str, error_details: str
) -> ComponentPage:
    return ComponentPage(
        name=convert_case(file_path.stem, "snake").replace("_", " ").title(),
        url_segment=convert_case(file_path.stem, "kebab").lower(),
        build=functools.partial(
            rio.components.error_placeholder.ErrorPlaceholder,
            error_summary,
            error_details,
        ),
    )



from __future__ import annotations

import io
import typing as t

import rio.data_models

# FIXME: Some explanations are nonsense, if an alignment is set. For example,
# setting grow to `True` does nothing if the component is aligned.

# FIXME: Specific explanations for
#
# - drawer
# - flow_container
# - grid
# - ListView
# - mouse_event_listener
# - popup
# - revealer
# - scroll_container
# - scroll_target


# These components pass on the entirety of the available space to their
# children
FULL_SIZE_SINGLE_CONTAINERS: set[type[rio.Component]] = {
    rio.Button,
    rio.Card,
    rio.Container,
    rio.CustomListItem,
    rio.KeyEventListener,
    rio.Link,
    rio.MouseEventListener,
    rio.PageView,
    rio.Rectangle,
    rio.Slideshow,
    rio.Stack,
    rio.Switcher,
}


# These components make use of the `grow_...` attributes in at least one axis.
CONTAINERS_SUPPORTING_GROW: set[type[rio.Component]] = {
    rio.Column,
    rio.Grid,
    rio.Row,
}


class LayoutExplainer:
    """
    Given a component, comes up with human readable explanations for the
    assigned layout.
    """

    session: rio.Session
    component: rio.Component
    _parent: rio.Component | None

    height_explanation: str
    width_explanation: str

    warnings: list[str]
    width_warnings: list[str]

    decrease_width: list[str]
    increase_width: list[str]

    decrease_height: list[str]
    increase_height: list[str]

    _layout: rio.data_models.ComponentLayout
    _parent_layout: rio.data_models.ComponentLayout | None

    def __init__(self) -> None:
        raise ValueError(
            "Don't create this class directly. Instead use the `create_new` method"
        )

    @staticmethod
    async def create_new(
        session: rio.Session,
        component: rio.Component,
    ) -> LayoutExplainer:
        """
        Creates and populates a new `LayoutExplainer` instance.

        ## Raises

        `KeyError`: If the given component cannot be found client-side.
        """
        # Create the instance
        self = LayoutExplainer.__new__(LayoutExplainer)
        self.session = session
        self.component = component

        self.warnings = []

        self.decrease_width = []
        self.increase_width = []

        self.decrease_height = []
        self.increase_height = []

        # Fetch layout information. This is asynchronous
        (self._layout,) = await session._get_component_layouts([component._id])

        if self._layout.parent_id is None:
            self._parent = None
            self._parent_layout = None
        else:
            self._parent = session._weak_components_by_id[
                self._layout.parent_id
            ]
            (self._parent_layout,) = await session._get_component_layouts(
                [self._layout.parent_id]
            )

        # Explain yourself!
        self.width_explanation = await self._explain_layout_in_axis(
            "width",
            suggest_shrink=self.decrease_width.append,
            suggest_grow=self.increase_width.append,
        )

        self.height_explanation = await self._explain_layout_in_axis(
            "height",
            suggest_shrink=self.decrease_height.append,
            suggest_grow=self.increase_height.append,
        )

        return self

    def _explain_allocated_space_before_alignment(
        self,
        axis_name: t.Literal["width", "height"],
        suggest_shrink: t.Callable[[str], None],
        suggest_grow: t.Callable[[str], None],
    ) -> str:
        """
        Given a component and its layout, return a human readable explanation
        for why the component was allocated the space it was. This doesn't
        account for alignment, but strictly explains the space handed down by
        the parent.
        """
        if axis_name == "width":
            allocated_space_before_alignment = (
                self._layout.allocated_outer_width
            )
            axis_xy = "x"
        else:
            allocated_space_before_alignment = (
                self._layout.allocated_outer_height
            )
            axis_xy = "y"

        target_class_name = type(self.component).__name__
        parent_class_name = type(self._parent).__name__

        # Try to get a specialized explanation based on the parent
        if self._parent_layout is None:
            return f"The component was allocated a {axis_name} of {allocated_space_before_alignment:.1f} by its parent."

        # Prepare some commonly used values
        if axis_name == "width":
            parent_allocated_size = self._parent_layout.allocated_inner_width
            parent_natural_size = self._parent_layout.natural_width
            is_grower = self.component.grow_x
        else:
            parent_allocated_size = self._parent_layout.allocated_inner_height
            parent_natural_size = self._parent_layout.natural_height
            is_grower = self.component.grow_y

        # Toplevel?
        if self.component is self.session._get_user_root_component():
            return f"This is the app's top-level component. As such, the {target_class_name} was allocated the full {axis_name} of {allocated_space_before_alignment:.1f} available in the window."

        # Single container?
        if type(self._parent) in FULL_SIZE_SINGLE_CONTAINERS:
            return f"Because `{parent_class_name}` components pass on all available space to their children, the component's {axis_name} is the full {allocated_space_before_alignment:.1f} units available in its parent."

        # Overlay
        if isinstance(self._parent, rio.Overlay):
            return f"Children of `Overlay` components aren't part of the normal layouting process. Instead, they're allocated the entire {axis_name} of the window, and can position themselves freely inside it using alignment. Because of this, the {target_class_name} was allocated the entire {allocated_space_before_alignment:.1f} units available in the window."

        # Dialog root elements
        if isinstance(self._parent, rio.DialogContainer):
            return f"Root components of dialogs are allocated the entire {axis_name} of the window, and can position themselves freely inside it using alignment. Because of this, the {target_class_name} was allocated the entire {allocated_space_before_alignment:.1f} units available in the window."

        # Row & Column
        if isinstance(self._parent, rio.Row) or isinstance(
            self._parent, rio.Column
        ):
            # Minor axis?
            if (
                isinstance(self._parent, rio.Row)
                and axis_name == "height"
                or isinstance(self._parent, rio.Column)
                and axis_name == "width"
            ):
                if is_grower:
                    self.warnings.append(
                        f"The component has `grow_{axis_xy}=True` set, but is placed inside of a `{parent_class_name}`. Because `{parent_class_name}`s pass on the entire available {axis_name} to all children it has no effect."
                    )

                return f"The component is placed inside of a `{parent_class_name}`. Since all children of `{parent_class_name}`s receive the full {axis_name}, it has received the entire {allocated_space_before_alignment:.1f} units available in its parent."

            # Major axis
            if self._parent.proportions is not None:
                suggest_shrink(
                    f"Adjust the `proportions` attribute of the parent to give less space to this {target_class_name}"
                )
                suggest_grow(
                    f"Adjust the `proportions` attribute of the parent to give more space to this {target_class_name}"
                )
                return f"The component is placed inside of a `{parent_class_name}`. Since the `{parent_class_name}` has a `proportions` attribute set, the available space was split up according to the proportions. Thus, this component was allocated a {axis_name} of {allocated_space_before_alignment:.1f}."

            # Only child
            if len(self._parent.children) == 1:
                return f"Because the component is the only child of its parent `{parent_class_name}`, it has received the full {allocated_space_before_alignment:.1f} units available in its parent."

            # No additional space
            if parent_allocated_size < parent_natural_size + 0.1:
                return f"The component is placed inside of a `{parent_class_name}`. Because that `{parent_class_name}` doesn't have any superfluous space available, the component was allocated the available {axis_name} of {allocated_space_before_alignment:.1f}."

            # Gather information about growers
            x_growers = 0
            y_growers = 0
            for child in self._parent.children:
                x_growers += child.grow_x
                y_growers += child.grow_y

            if axis_name == "width":
                n_growers = x_growers
                is_grower = self.component.grow_x
            else:
                n_growers = y_growers
                is_grower = self.component.grow_y

            shrink_by_growing_others = f"Assign `grow_{axis_xy}=True` to one of the other children of the `{parent_class_name}` component, so it takes up the superfluous space"
            grow_by_growing = f"Assign `grow_{axis_xy}=True` to the component, so it is preferentially assigned the superfluous space"
            grow_by_shrinking_others = f"Remove the `grow_{axis_xy}=True` from one of the other children of the `{parent_class_name}` component, so this component gets more of the superfluous space"

            # No growers
            if n_growers == 0:
                suggest_shrink(shrink_by_growing_others)
                suggest_grow(grow_by_growing)
                return f"The component is placed inside of a `{parent_class_name}`. Because none of the `{parent_class_name}`'s children are set to grow, the `{parent_class_name}` has split up the superfluous space evenly between all children. Thus, this component has received a {axis_name} of {allocated_space_before_alignment:.1f}."

            # There are growers, but this is not one of them
            if not is_grower:
                suggest_grow(grow_by_growing)
                return f"The component is placed inside of a `{parent_class_name}`. Because the component is not set to grow, the superfluous space was given to other children. Thus, this component was left with its minimum {axis_name} of {allocated_space_before_alignment:.1f}."

            # Multiple growers, including this one
            if n_growers > 1:
                suggest_shrink(shrink_by_growing_others)
                suggest_grow(grow_by_shrinking_others)
                return f"The component is placed inside of a `{parent_class_name}`. The superfluous space was split up evenly between all children with a `grow_{axis_xy}=True` attribute. Thus, this component was allocated a {axis_name} of {allocated_space_before_alignment:.1f}."

            # Only grower
            suggest_shrink(shrink_by_growing_others)
            return f"The component is placed inside of a `{parent_class_name}`. Because it is the only child of the `{parent_class_name}` with a `grow_{axis_xy}=True` attribute, it has received all of the superfluous space available in its parent. Thus, it was allocated a {axis_name} of {allocated_space_before_alignment:.1f}."

        # No specialized explanation is available. Fall back to a generic default
        return f"The component was allocated a {axis_name} of {allocated_space_before_alignment:.1f} by its parent `{parent_class_name}`."

    async def _explain_layout_in_axis(
        self,
        axis_name: t.Literal["width", "height"],
        suggest_shrink: t.Callable[[str], None],
        suggest_grow: t.Callable[[str], None],
    ) -> str:
        """
        Given a component, come up with a human readable explanation for why it was
        allocated the space it was, in a single direction.
        """
        # Prepare some values based on axis
        if axis_name == "width":
            allocated_size = self._layout.allocated_inner_width
            allocated_size_before_alignment = self._layout.allocated_outer_width
            specified_min_size = self.component.min_width
            natural_size = self._layout.natural_width
            total_margin = (
                self.component._effective_margin_left_
                + self.component._effective_margin_right_
            )
            axis_xy = "x"
            start = "left"
            end = "right"
            alignment = self.component.align_x
            is_grower = self.component.grow_x
        else:
            allocated_size = self._layout.allocated_inner_height
            allocated_size_before_alignment = (
                self._layout.allocated_outer_height
            )
            specified_min_size = self.component.min_height
            natural_size = self._layout.natural_height
            total_margin = (
                self.component._effective_margin_top_
                + self.component._effective_margin_bottom_
            )
            axis_xy = "y"
            start = "top"
            end = "bottom"
            alignment = self.component.align_y
            is_grower = self.component.grow_y

        target_class_name = type(self.component).__name__
        parent_class_name = type(self._parent).__name__

        # Warn if the component has a `grow` attribute set, but is placed inside
        # of a container that doesn't support it
        if (
            is_grower
            and self._parent is not None
            and type(self._parent) not in CONTAINERS_SUPPORTING_GROW
        ):
            self.warnings.append(
                f"The component has `grow_{axis_xy}=True` set, but is placed inside of a `{parent_class_name}`. {parent_class_name} components do not make use of this property, so it has no effect."
            )

        # Warn if the component is aligned, but has no natural size
        if alignment is not None and natural_size < 0.1:
            self.warnings.append(
                f"The component is aligned using `align_{axis_xy}` but it has no natural {axis_name}. Since aligned components receive the minimum amount of space necessary, and this component doesn't require any space, it will not be visible."
            )

        # How much space did the parent hand down?
        result = io.StringIO()
        result.write(
            self._explain_allocated_space_before_alignment(
                axis_name,
                suggest_shrink=suggest_shrink,
                suggest_grow=suggest_grow,
            )
        )
        result.write(" ")

        # Alignment
        if allocated_size_before_alignment < natural_size + total_margin + 0.1:
            result.write(
                f"This matches the {axis_name} needed by the component."
            )
        elif alignment is None:
            result.write(
                "Because no alignment is set, it uses all of that space."
            )
        else:
            result.write("\n\n")
            result.write(
                f"Due to `align_{axis_xy}` being set, the {target_class_name} only takes up the minimum amount of space necessary"
            )

            if alignment <= 0.03:
                result.write(
                    f" and is located at the {start} of the available space."
                )
            elif 0.47 <= alignment <= 0.53:
                result.write(f" and is centered in the available space.")
            elif alignment >= 0.97:
                result.write(
                    f" and is located at the {end} of the available space."
                )
            else:
                result.write(
                    f", with {alignment * 100:.0f}% of the leftover space on the {start}, and the remainder on the {end}."
                )

        # Warn if the specified minimum size is less than the natural one
        if 0 < specified_min_size < natural_size:
            self.warnings.append(
                f"The explicitly set minimum {axis_name} of {specified_min_size:.1f} has no effect, because it is less than the component's natural {axis_name} of {natural_size:.1f}. Components can never be smaller than their natural size."
            )

        # Suggest growing the component by setting an explicit size
        if allocated_size == 0:
            suggest_grow(f"Assign a `min_{axis_name}` to the component")
        else:
            suggest_grow(
                f"Assign a `min_{axis_name}` to the component that is greater than its allocated {axis_name} of {allocated_size:.1f}"
            )

        # If the component isn't aligned, suggest to grow it by growing its
        # parent
        if alignment is None:
            suggest_grow(
                f"Increase the {axis_name} of the parent component, so it hands down more space to the {target_class_name}"
            )

        # If the component has more space available than it needs, suggest
        # using an alignment to shrink
        if alignment is None and allocated_size > natural_size + 0.1:
            suggest_shrink(
                f"Align the component using `align_{axis_xy}`, so it only takes up its natural {axis_name}"
            )

        # If aligned and the minimum size exceeds the natural size, suggest
        # removing the minimum
        if alignment is not None and specified_min_size > natural_size:
            suggest_shrink(
                f"Remove the `min_{axis_name}` attribute from the component so it only takes up its natural {axis_name}"
            )

        # If aligned, the component can be grown by removing the alignment
        if (
            alignment is not None
            and (allocated_size + 0.1) < allocated_size_before_alignment
        ):
            suggest_grow(
                f"Remove the `align_{axis_xy}` attribute from the component, so it takes up all of the available space"
            )

        # If the component is at its minimum size, give tips on how to reduce
        # its minimum size further
        if allocated_size < natural_size + 0.1:
            for suggestion in self._explain_how_to_reduce_natural_size(
                axis_name
            ):
                suggest_shrink(suggestion)

        # Done!
        return result.getvalue()

    def _explain_how_to_reduce_natural_size(
        self, axis_name: t.Literal["width", "height"]
    ) -> t.Iterable[str]:
        if isinstance(self.component, rio.Text):
            if axis_name == "width":
                yield 'Set `overflow="wrap"` or `overflow="ellipsize"` to reduce the `Text`\'s natural width'



import contextlib
import pathlib
import subprocess
import sys
import typing as t
from pathlib import Path

import imy.docstrings
import introspection.types
import introspection.typing
import pandas
import polars

import rio.docs

ALIASES: dict[object, str] = {
    obj: f"rio.{doc.name}"
    for obj, doc in rio.docs.get_documented_objects().items()
}
ALIASES[pandas.DataFrame] = "pandas.DataFrame"
ALIASES[polars.DataFrame] = "polars.DataFrame"


def main():
    stub_file_path = Path(rio.__file__).absolute().parent / "__init__temp.pyi"

    with stub_file_path.open("w", encoding="utf8") as stub_file:
        writer = StubWriter(stub_file)

        for docs in rio.docs.get_documented_objects().values():
            writer.write(docs)

    # Run the file through a linter to ensure its correctness
    subprocess.run(
        [sys.executable, "-m", "ruff", "check", str(stub_file_path)],
        check=True,
    )

    subprocess.run(
        [sys.executable, "-m", "ruff", "format", str(stub_file_path)],
        check=True,
    )


class StubWriter:
    def __init__(self, file: t.TextIO):
        self._file = file
        self._indent = 0

        self._write("""
import datetime
import os
import pathlib
import typing

import numpy
import pandas
import polars
import PIL.Image
import yarl

import rio


""")

    @contextlib.contextmanager
    def _indented(self):
        self._indent += 1
        yield
        self._indent -= 1

    def _write_line(self, line: str) -> None:
        self._write_indented(line)
        self._write("\n")

    def _write_indented(self, text: str) -> None:
        self._write("    " * self._indent)
        self._write(text)

    def _write(self, text: str) -> None:
        self._file.write(text)

    def write(
        self, docs: imy.docstrings.FunctionDocs | imy.docstrings.ClassDocs
    ) -> None:
        if docs.object is rio.URL:
            self._write_line(f"{docs.name} = yarl.URL")
            return

        if isinstance(docs, imy.docstrings.FunctionDocs):
            self._write_function(docs)
        else:
            self._write_class(docs)

    def _write_function(self, docs: imy.docstrings.FunctionDocs) -> None:
        if docs.class_method:
            self._write_line("@classmethod")
        elif docs.static_method:
            self._write_line("@staticmethod")

        self._write_function_signature(docs)

        with self._indented():
            if docs.summary or docs.details:
                self._write_docstring(docs)
            else:
                self._write_line("...")

        self._write("\n")

    def _write_function_signature(
        self, docs: imy.docstrings.FunctionDocs
    ) -> None:
        if not docs.parameters:
            self._write_line(f"def {docs.name}():")
            return

        self._write_line(f"def {docs.name}(")

        with self._indented():
            already_kw_only = False

            for parameter in docs.parameters:
                if parameter.collect_positional:
                    name = "*" + parameter.name
                    already_kw_only = True
                elif parameter.collect_keyword:
                    name = "**" + parameter.name
                    already_kw_only = True
                elif parameter.kw_only and not already_kw_only:
                    self._write_line("*,")
                    name = parameter.name
                    already_kw_only = True
                else:
                    name = parameter.name

                self._write_annotated_name(
                    name, parameter.type, parameter.default, end=",\n"
                )

        self._write_line("):")

    def _write_annotated_name(
        self,
        name: str,
        type_: introspection.types.TypeAnnotation | imy.docstrings.Unset,
        default_value: object | imy.docstrings.Unset = imy.docstrings.UNSET,
        *,
        end: str = "\n",
        force_annotation: bool = False,
    ) -> None:
        self._write_indented(name)

        if isinstance(type_, imy.docstrings.Unset):
            if force_annotation:
                self._write(": typing.Any")
        else:
            self._write(": ")
            self._write(
                introspection.typing.annotation_to_string(
                    type_, implicit_typing=False, aliases=ALIASES
                )
            )

        if not isinstance(default_value, imy.docstrings.Unset):
            self._write(f" = {repr_value(default_value)}")

        self._write(end)

    def _write_docstring(
        self, docs: imy.docstrings.FunctionDocs | imy.docstrings.ClassDocs
    ) -> None:
        docstring = "\n\n".join(filter(None, [docs.summary, docs.details]))

        if not docstring:
            return

        self._write_line(repr(docstring))
        self._write("\n")

    def _write_class(self, docs: imy.docstrings.ClassDocs) -> None:
        if docs.object is not rio.Component:
            self._write_line("@typing.final")

        self._write_indented(f"class {docs.name}")
        self._write_base_classes(docs)
        self._write(":\n")

        with self._indented():
            self._write_docstring(docs)

            for attr in docs.attributes:
                self._write_annotated_name(attr.name, attr.type)

            if docs.attributes:
                self._write("\n")

            for method in docs.functions:
                self._write_function(method)

        self._write("\n")

    def _write_base_classes(self, docs: imy.docstrings.ClassDocs) -> None:
        public_base_classes = [
            rio.docs.get_documented_objects()[cls]
            for cls in docs.object.__bases__
            if cls in rio.docs.get_documented_objects()
        ]

        if not public_base_classes:
            return

        self._write("(")
        self._write(", ".join(docs.name for docs in public_base_classes))
        self._write(")")


def repr_value(value: object) -> str:
    if isinstance(value, (pathlib.WindowsPath, pathlib.PosixPath)):
        return f"pathlib.Path({value})"

    value_repr = repr(value)

    # Some objects, like functions, don't have a proper repr. Replace
    # those values with "...".
    if value_repr.startswith("<"):
        return "..."

    # Replace occurrences of "WindowsPath" and "PosixPath". (This is done as a
    # string operation because path objects can appear in nested data
    # structures. We don't want to recurse into every container just to call
    # `isinstance(..., (WindowsPath, PosixPath))` on every object.)
    value_repr = value_repr.replace("WindowsPath(", "pathlib.Path(")
    value_repr = value_repr.replace("PosixPath(", "pathlib.Path(")

    # Remove type annotations for sentinel default values
    value_repr = value_repr.replace(" | rio.text_style.UnsetType", "")

    return value_repr


if __name__ == "__main__":
    main()



from __future__ import annotations

import copy
import dataclasses
import json
import typing as t

import typing_extensions as te
import uniserde

import rio

from .snippet_manager import Snippet

__all__ = [
    "ProjectTemplate",
    "get_project_templates",
]


# Contains default values for the `meta.json` file
DEFAULT_META_DICT = {
    "dependencies": {},
    "rootComponent": None,
    "onAppStart": None,
    "onSessionStart": None,
    "defaultAttachments": None,
    "theme": None,
}


# All Available templates, including the empty template
#
# THE ORDER MATTERS. `revel` will display the options in the same order as they
# appear in the literal
AvailableTemplatesLiteral: te.TypeAlias = t.Literal[
    # Keep the empty template first
    "Empty",
    # Sort the remainder alphabetically
    "AI Chatbot",
    "Authentication",
    "Crypto Dashboard",
    "Multipage Website",
    "Simple CRUD",
    "Tic-Tac-Toe",
    "Todo App",
]


@dataclasses.dataclass
class _TemplateConfig(uniserde.Serde):
    """
    Model for parsing the JSON file which comes along with each project
    template.
    """

    # Allows displaying the templates in a structured way
    level: t.Literal["beginner", "intermediate", "advanced"]

    # Very short, one or two line description of the template
    summary: str

    # Any pypi packages this project depends on, in the same format as used by
    # `pip`.
    #
    # Example: `{"numpy": ">=1.21.0", "pandas": "^1.3.0"}`
    dependencies: dict[str, str]

    # Whether projects based on this template are ready to run out of the box,
    # without any modifications needed. For example, a template that requires an
    # API key to be set up is not ready to run.
    ready_to_run: bool

    # Additional parameters to pass to the app instance
    root_component: str | None
    on_app_start: str | None
    on_session_start: str | None
    default_attachments: list[str] | None
    theme: str | None


@dataclasses.dataclass
class ProjectTemplate:
    """
    Project templates are stored as snippets. This class represents all
    resources necessary for a single project template.
    """

    # Human-readable name of the project template
    name: AvailableTemplatesLiteral

    # How difficult the project is
    level: t.Literal["beginner", "intermediate", "advanced"]

    # A short description of the project template
    summary: str

    # Description of the project template (Markdown)
    description_markdown_source: str

    # The project thumbnail. This is a SVG file.
    thumbnail: Snippet

    # Any pypi packages this project depends on, in the same format as used by
    # `pip`.
    #
    # Example: `{"numpy": ">=1.21.0", "pandas": "^1.3.0"}`
    dependencies: dict[str, str]

    # Whether projects based on this template are ready to run out of the box,
    # without any modifications needed. For example, a template that requires an
    # API key to be set up is not ready to run.
    ready_to_run: bool

    # All snippets which should be included
    asset_snippets: list[Snippet]
    component_snippets: list[Snippet]
    page_snippets: list[Snippet]
    other_python_files: list[Snippet]

    root_init_snippet: Snippet

    # Additional configuration for the app instance
    root_component: str | None
    on_app_start: str | None
    on_session_start: str | None
    default_attachments: list[str] | None
    theme: str | None = None

    @property
    def slug(self) -> str:
        return self.name.lower().replace(" ", "-")

    @property
    def live_example_url(self) -> rio.URL | None:
        """
        If there is a publicly hosted demo for this template, return the URL to
        it. Otherwise, return `None`.
        """
        # All examples are hosted, provided they don't require any changes by
        # the user.
        if self.ready_to_run:
            return rio.URL(f"https://rio.dev/live-example/{self.slug}")

        return None

    @property
    def source_code_url(self) -> rio.URL:
        """
        All built-in templates have their source code publicly available on GitHub. This
        returns the URL to the repository.
        """
        return rio.URL(
            f"https://github.com/rio-labs/rio-templates/tree/main/{self.slug}"
        )

    @staticmethod
    def _from_snippet_group(
        snippet_name: str,
        snippets: t.Iterable[Snippet],
    ) -> ProjectTemplate:
        assert (
            snippet_name in t.get_args(AvailableTemplatesLiteral)
            or snippet_name == "Empty"
        ), snippet_name
        name = t.cast(AvailableTemplatesLiteral, snippet_name)

        # Find all snippets needed for the project template
        readme_snippet: Snippet | None = None
        thumbnail_snippet: Snippet | None = None
        metadata: _TemplateConfig | None = None
        root_init_snippet: Snippet | None = None

        asset_snippets: list[Snippet] = []
        python_files: list[Snippet] = []

        for snippet in snippets:
            # README snippet can be recognized by its name
            if snippet.name == "README.md":
                assert readme_snippet is None
                assert snippet.is_text_snippet, snippet.file_path
                readme_snippet = snippet
                continue

            # As is the thumbnail
            if snippet.name == "thumbnail.svg":
                assert thumbnail_snippet is None
                thumbnail_snippet = snippet
                continue

            # And the metadata
            if snippet.name == "meta.json":
                meta_dict: dict[str, t.Any] = copy.deepcopy(DEFAULT_META_DICT)
                meta_dict.update(json.loads(snippet.stripped_code()))
                metadata = _TemplateConfig.from_json(meta_dict)
                continue

            # Others are categorized by the directory they're in
            dir_name = snippet.file_path.parent.name

            if dir_name == "assets":
                asset_snippets.append(snippet)
            # if "assets" in snippet.file_path.parts:
            #     asset_snippets.append(snippet)

            elif snippet.file_path.name == "root_init.py":
                assert root_init_snippet is None
                assert snippet.is_text_snippet, snippet.file_path
                root_init_snippet = snippet

            elif snippet.file_path.suffix == ".py":
                python_files.append(snippet)

            elif snippet.file_path.suffix in [
                ".jpg",
                ".png",
                ".jpeg",
                ".svg",
                ".webp",
            ]:
                asset_snippets.append(snippet)
            # else:
            #     assert False, f"Unrecognized snippet file `{snippet.file_path}`"

            else:
                assert False, f"Unrecognized snippet file `{snippet.file_path}`"

        # Make sure everything was found
        assert readme_snippet is not None, (
            f"`README.md` snippet not found for `{name}`"
        )
        assert thumbnail_snippet is not None, (
            f"`thumbnail.svg` snippet not found for {name}"
        )
        assert metadata is not None, (
            f"`meta.json` snippet not found for `{name}`"
        )
        assert root_init_snippet is not None, (
            f"`root_init.py` snippet not found for `{name}`"
        )

        # Further split the Python files into components, pages, and other
        # files. This cannot be done above, because it requires knowledge of
        # where the snippet groups's root directory is placed. This directory in
        # turn is only known once the snippets have been found.
        template_root_dir = root_init_snippet.file_path.parent

        ii = 0
        component_snippets: list[Snippet] = []
        page_snippets: list[Snippet] = []

        while ii < len(python_files):
            snippet = python_files[ii]
            rel_path = snippet.file_path.relative_to(template_root_dir)

            # Component?
            if rel_path.parts[0] == "components":
                del python_files[ii]

                if snippet.name == "__init__.py":
                    continue

                assert snippet.is_text_snippet, snippet.file_path
                component_snippets.append(snippet)

            # Page?
            elif rel_path.parts[0] == "pages":
                del python_files[ii]

                assert snippet.is_text_snippet, snippet.file_path
                page_snippets.append(snippet)

            # Plain old Python file
            else:
                ii += 1

        assert len(page_snippets) > 0, f"No page snippets found for `{name}`"

        # Create the project template
        return ProjectTemplate(
            name=name,
            level=metadata.level,
            summary=metadata.summary,
            description_markdown_source=readme_snippet.stripped_code(),
            thumbnail=thumbnail_snippet,
            dependencies=metadata.dependencies,
            asset_snippets=asset_snippets,
            component_snippets=component_snippets,
            page_snippets=page_snippets,
            other_python_files=python_files,
            root_init_snippet=root_init_snippet,
            root_component=metadata.root_component,
            on_app_start=metadata.on_app_start,
            on_session_start=metadata.on_session_start,
            default_attachments=metadata.default_attachments,
            theme=metadata.theme,
            ready_to_run=metadata.ready_to_run,
        )



from __future__ import annotations

import abc
import asyncio
import inspect
import json
import logging
import time
import traceback
import warnings
import weakref
from datetime import date
from pathlib import Path

import langcodes
import pytz
import starlette.datastructures

import rio
import rio.assets

from .. import (
    assets,
    data_models,
    language_info,
    routing,
    session,
    user_settings_module,
    utils,
)
from ..transports import (
    AbstractTransport,
    TransportClosedIntentionally,
    TransportInterrupted,
)

__all__ = ["AbstractAppServer"]


class AbstractAppServer(abc.ABC):
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool,
        debug_mode: bool,
        base_url: rio.URL | None = None,
    ) -> None:
        self.app = app
        self.running_in_window = running_in_window
        self.debug_mode = debug_mode

        # This is the URL the app's home page is hosted at, as seen from the
        # client. So if the user needs to type `https://example.com/my-app/` to
        # see the app, this will be `https://example.com/my-app/`. Note that
        # this is not necessarily the URL Python gets to see in HTTP requests,
        # as it is common practice to have a reverse proxy rewrite the URLs of
        # HTTP requests.
        #
        # An important consideration with the base URL is, that the full URL
        # must be known for some scenarios. For example, the `robots.txt` has to
        # return the full URL of the sitemap, since not all (none?) search
        # engines will follow relative URLs.
        #
        # This is currently handled as follows: If a base URL is specified, it
        # must contain the full URL. If no base URL is specified, the URL
        # provided in the request is used instead.
        self.base_url = base_url
        self._session_serve_tasks = dict[rio.Session, asyncio.Task[object]]()
        self._disconnected_sessions = dict[rio.Session, float]()

    @property
    def sessions(self) -> list[rio.Session]:
        return list(self._session_serve_tasks)

    async def _on_start(self) -> None:
        # Trigger the app's startup event
        #
        # This will be done blockingly, so the user can prepare any state before
        # connections are accepted. This is also why it's called before the
        # internal event.
        await self._call_on_app_starts()

        # If running as a server, periodically clean up expired sessions
        if not self.running_in_window:
            asyncio.create_task(
                _periodically_clean_up_expired_sessions(weakref.ref(self)),
                name="Periodic session cleanup",
            )

    async def _on_close(self) -> None:
        # Close all sessions
        rio._logger.debug(
            f"App server shutting down; closing {len(self.sessions)} active session(s)"
        )

        # On Linux, some of these tasks are sometimes canceled, resulting in a
        # visible, ugly traceback in the console. Debugging has been fruitless
        # so far, hence the `asyncio.shield` hack.
        #
        # That wasn't enough though - hence the additional `try` block, because
        # nothing beats fixing a bad bandage than adding another one.
        try:
            results = await asyncio.shield(
                asyncio.gather(
                    *(
                        sess._close(close_remote_session=True)
                        for sess in self.sessions
                    ),
                    return_exceptions=True,
                )
            )
        except asyncio.CancelledError:
            pass
        else:
            for result in results:
                if isinstance(result, BaseException):
                    traceback.print_exception(
                        type(result), result, result.__traceback__
                    )

        await self._call_on_app_close()

    async def _call_on_app_starts(self) -> None:
        rio._logger.debug("Calling `on_app_start`")

        # Call the app's function
        if self.app._on_app_start is not None:
            try:
                result = self.app._on_app_start(self.app)

                if inspect.isawaitable(result):
                    await result

            # Display and discard exceptions
            except Exception:
                print("Exception in `on_app_start` event handler:")
                traceback.print_exc()

        # Call the functions of any extensions
        await self.app._call_event_handlers(
            self.app._extension_on_app_start_handlers,
            rio.ExtensionAppStartEvent(
                self.app,
            ),
        )

    async def _call_on_app_close(self) -> None:
        rio._logger.debug("Calling `on_app_close`")

        # Call the app's function
        if self.app._on_app_close is not None:
            try:
                result = self.app._on_app_close(self.app)

                if inspect.isawaitable(result):
                    await result

            # Display and discard exceptions
            except Exception:
                print("Exception in `_on_app_close` event handler:")
                traceback.print_exc()

        # Call the functions of any extensions
        await self.app._call_event_handlers(
            self.app._extension_on_app_close_handlers,
            rio.ExtensionAppCloseEvent(
                self.app,
            ),
        )

    @abc.abstractmethod
    async def pick_file(
        self,
        session: rio.Session,
        *,
        file_types: list[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | list[utils.FileInfo]:
        raise NotImplementedError

    @abc.abstractmethod
    def external_url_for_user_asset(self, relative_asset_path: Path) -> rio.URL:
        """
        Returns the URL of an asset from the app's `assets_dir`. This must be a
        permalink, i.e. a URL that doesn't change if the server is restarted.

        The returned URL is **external**, i.e. which value would have to be
        typed into the browser to access it. The URL visible to Python can
        differ from this, e.g. because of a reverse proxy.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def weakly_host_asset(self, asset: assets.HostedAsset) -> rio.URL:
        """
        Hosts an asset as long as it is alive. Returns the asset's URL as a
        string.

        The returned URL is **external**, i.e. which value would have to be
        typed into the browser to access it. The URL visible to Python can
        differ from this, e.g. because of a reverse proxy.
        """
        raise NotImplementedError

    def host_asset_with_timeout(
        self, asset: assets.HostedAsset, timeout: float
    ) -> rio.URL:
        """
        Hosts an asset for a limited time. Returns the asset's URL as a string.
        """
        url = self.weakly_host_asset(asset)

        async def keep_alive() -> None:
            await asyncio.sleep(timeout)
            _ = asset

        asyncio.create_task(
            keep_alive(), name=f"Keep asset {asset} alive for {timeout} sec"
        )

        return url

    def _after_session_closed(self, session: rio.Session) -> None:
        """
        Called by `Session.close()`. Gives the server an opportunity to clean
        up.
        """
        # Stop the task that's listening for incoming messages
        task = self._session_serve_tasks.pop(session)
        task.cancel("Session has closed")

    async def create_session(
        self,
        initial_message: data_models.InitialClientMessage,
        *,
        transport: AbstractTransport,
        client_ip: str,
        client_port: int,
        http_headers: starlette.datastructures.Headers,
    ) -> rio.Session:
        """
        Creates a new session.

        ## Raises

        `NavigationFailed`: If a page guard crashes.

        `NavigationFailed`: If the initial page URL is not a child of the app's
            base URL.
        """
        # Normalize and deduplicate the languages
        preferred_languages: list[str] = []

        for language in initial_message.preferred_languages:
            try:
                language = langcodes.standardize_tag(language)
            except ValueError:
                continue

            if language not in preferred_languages:
                preferred_languages.append(language)

        if len(preferred_languages) == 0:
            preferred_languages.append("en-US")

        # Get locale information
        if len(initial_message.decimal_separator) != 1:
            logging.warning(
                f'Client sent invalid decimal separator "{initial_message.decimal_separator}". Using "." instead.'
            )
            initial_message.decimal_separator = "."

        if len(initial_message.thousands_separator) > 1:
            logging.warning(
                f'Client sent invalid thousands separator "{initial_message.thousands_separator}". Using "" instead.'
            )
            initial_message.thousands_separator = ""

        # There does not seem to be any good way to determine the first day of
        # the week in JavaScript. Look it up based on the preferred language.
        first_day_of_week = language_info.get_week_start_day(
            preferred_languages[0]
        )

        # Make sure the date format string is valid
        try:
            formatted_date = date(3333, 11, 22).strftime(
                initial_message.date_format_string
            )
        except ValueError:
            date_format_string_is_valid = False
        else:
            date_format_string_is_valid = (
                "33" in formatted_date
                and "11" in formatted_date
                and "22" in formatted_date
            )

        if not date_format_string_is_valid:
            logging.warning(
                f'Client sent invalid date format string "{initial_message.date_format_string}". Using "%Y-%m-%d" instead.'
            )
            initial_message.date_format_string = "%Y-%m-%d"

        # Parse the timezone
        try:
            timezone = pytz.timezone(initial_message.timezone)
        except pytz.UnknownTimeZoneError:
            logging.warning(
                f'Client sent unknown timezone "{initial_message.timezone}". Using UTC instead.'
            )
            timezone = pytz.UTC

        # Set the theme according to the user's preferences
        theme = self.app._theme
        if isinstance(theme, tuple):
            if initial_message.prefers_light_theme:
                theme = theme[0]
            else:
                theme = theme[1]

        # Prepare the initial URL. This will be exposed to the session as the
        # `active_page_url`, but overridden later once the page guards have been
        # run.
        initial_page_url = rio.URL(initial_message.url)

        # Determine the base URL. If one was explicitly provided when creating
        # the app server, use that. Otherwise derive one from the initial HTTP
        # request.
        if self.base_url is None:
            base_url = (
                initial_page_url.with_path("").with_query("").with_fragment("")
            )
        else:
            base_url = self.base_url

        # Create the session
        sess = session.Session(
            app_server_=self,
            transport=transport,
            client_ip=client_ip,
            client_port=client_port,
            http_headers=http_headers,
            base_url=base_url,
            active_page_url=initial_page_url,
            timezone=timezone,
            preferred_languages=preferred_languages,
            month_names_long=initial_message.month_names_long,
            day_names_long=initial_message.day_names_long,
            date_format_string=initial_message.date_format_string,
            first_day_of_week=first_day_of_week,
            decimal_separator=initial_message.decimal_separator,
            thousands_separator=initial_message.thousands_separator,
            screen_width=initial_message.screen_width,
            screen_height=initial_message.screen_height,
            window_width=initial_message.window_width,
            window_height=initial_message.window_height,
            pixels_per_font_height=initial_message.physical_pixels_per_font_height,
            scroll_bar_size=initial_message.scroll_bar_size,
            primary_pointer_type=initial_message.primary_pointer_type,
            theme_=theme,
        )

        # Deserialize the user settings
        await sess._load_user_settings(initial_message.user_settings)

        # Add any remaining attachments
        for attachment in self.app.default_attachments:
            if not isinstance(attachment, user_settings_module.UserSettings):
                sess.attach(attachment)

        # Start listening for incoming messages. This should happen before
        # `on_session_start` is called, so that we don't deadlock in case
        # someone calls a method that requires a response from the client.
        self._session_serve_tasks[sess] = asyncio.create_task(
            self._serve_session(sess),
            name=f"`Session.serve` for session id `{id(sess)}`",
        )

        # Trigger the `on_session_start` event.
        #
        # Since this event is often used for important initialization tasks like
        # adding attachments, actually wait for it to finish before continuing.
        #
        # However, also don't run it too early, since this function expects a
        # (mostly) functioning session.
        #
        # TODO: Figure out which values are still missing, and expose them,
        #       expose placeholders, or document that they aren't available.
        start_time = time.monotonic()
        await sess._call_event_handler(
            self.app._on_session_start,
            sess,
            refresh=False,
        )
        duration = time.monotonic() - start_time

        if duration > 5:
            warnings.warn(
                f"Session startup was delayed for {duration:.0f} seconds by"
                f" `on_session_start`. If you have long-running operations that"
                f" don't have to finish before the session starts, you should"
                f" execute them in a background task."
            )

        # Extensions may also have `on_session_start` handlers
        await self.app._call_event_handlers(
            self.app._extension_on_session_start_handlers,
            rio.ExtensionSessionStartEvent(
                sess,
            ),
        )

        # Run any page guards for the initial page. Throws a `NavigationFailed`
        # if a page guard crashed.
        (
            active_page_instances_and_path_arguments,
            active_page_url_absolute,
        ) = routing.check_page_guards(sess, initial_page_url)

        # The initial page URL can be problematic. Logically it must obviously
        # be a child of the app's base URL, but that may not be the case if a
        # malicious URL was sent or the base misconfigured. Check for this.
        if active_page_instances_and_path_arguments is None:
            raise rio.NavigationFailed(
                f"The initial page URL `{initial_page_url}` is not a child of the app's base URL `{self.base_url}`."
            )

        # Is this a page, or a full URL to another site?
        try:
            utils.url_relative_to_base(
                sess._base_url,
                active_page_url_absolute,
            )

        # This is an external URL. Navigate to it
        except ValueError:

            async def history_worker() -> None:
                await sess._evaluate_javascript(
                    f"""
                    window.location.href = {json.dumps(str(active_page_url_absolute))};
                    """
                )

            sess.create_task(history_worker(), name="navigate to external URL")

            # TODO: End the session? Abort initialization?

        # Set the initial page URL. When connecting to the server, all relevant
        # page guards execute. These may change the URL of the page, so the
        # client needs to take care to update the browser's URL to match the
        # server's.
        if active_page_url_absolute != initial_page_url:

            async def update_url_worker():
                js_page_url = json.dumps(str(active_page_url_absolute))
                await sess._evaluate_javascript(
                    f"""
                    console.trace("Updating browser URL to match the one modified by guards:", {js_page_url});
                    window.history.replaceState(null, "", {js_page_url});
                    """
                )

            sess.create_task(
                update_url_worker(),
                name="Update browser URL to match the one modified by guards",
            )

        # Update the session's active page and instances
        sess._active_page_url = active_page_url_absolute
        sess._active_page_instances_and_path_arguments = (
            active_page_instances_and_path_arguments
        )
        sess._active_page_instances = tuple(
            page for page, _ in active_page_instances_and_path_arguments
        )

        # Apply the CSS for the chosen theme
        await sess._apply_theme(theme)

        # Send the first `updateComponentStates` message
        await sess._refresh()

        return sess

    async def _serve_session(self, sess: rio.Session) -> None:
        try:
            await sess.serve()
        except TransportClosedIntentionally:
            sess.close()
        except TransportInterrupted:
            # Connection was interrupted, mark the session as disconnected but
            # keep it alive for a while to see if the client reconnects
            sess._transport = None


async def _periodically_clean_up_expired_sessions(
    app_server_ref: weakref.ReferenceType[AbstractAppServer],
) -> None:
    LOOP_INTERVAL = 60 * 15
    SESSION_LIFETIME = 60 * 60

    while True:
        await asyncio.sleep(LOOP_INTERVAL)

        app_server = app_server_ref()
        if app_server is None:
            return

        now = time.monotonic()
        cutoff = now - SESSION_LIFETIME

        disconnected_sessions = list(app_server._disconnected_sessions.items())
        for sess, disconnect_time in disconnected_sessions:
            if disconnect_time < cutoff:
                sess.close()

        # Drop the reference to the app server
        app_server = None



import rio.testing


async def test_fundamental_container_as_root() -> None:
    def build() -> rio.Component:
        return rio.Row(rio.Text("Hello"))

    async with rio.testing.TestClient(build) as test_client:
        row_component = test_client.get_component(rio.Row)
        text_component = test_client.get_component(rio.Text)

        assert {
            row_component,
            text_component,
        } <= test_client._last_updated_components



import asyncio
import typing as t

import ordered_set
import starlette.datastructures
import typing_extensions as te
from uniserde import JsonDoc

import rio

from . import data_models
from .app_server import TestingServer
from .transports import MessageRecorderTransport, TransportInterrupted

__all__ = ["TestClient"]


T = t.TypeVar("T")
C = t.TypeVar("C", bound=rio.Component)


class TestClient:
    @t.overload
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ): ...

    @t.overload
    def __init__(
        self,
        build: t.Callable[[], rio.Component] = rio.Spacer,
        *,
        app_name: str = "mock-app",
        default_attachments: t.Iterable[object] = (),
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ): ...

    def __init__(  # type: ignore
        self,
        app_or_build: rio.App | t.Callable[[], rio.Component] | None = None,
        *,
        app: rio.App | None = None,
        build: t.Callable[[], rio.Component] | None = None,
        app_name: str = "test-app",
        default_attachments: t.Iterable[object] = (),
        running_in_window: bool = False,
        user_settings: JsonDoc = {},
        active_url: str = "/",
        use_ordered_dirty_set: bool = False,
    ):
        if app is None:
            if isinstance(app_or_build, rio.App):
                app = app_or_build
            else:
                if build is None:
                    if app_or_build is not None:
                        build = app_or_build
                    else:
                        build = rio.Spacer

                app = rio.App(
                    build=build,
                    name=app_name,
                    default_attachments=tuple(default_attachments),
                )

        self._app_server = TestingServer(
            app,
            debug_mode=False,
            running_in_window=running_in_window,
        )

        self._user_settings = user_settings
        self._active_url = active_url
        self._use_ordered_dirty_set = use_ordered_dirty_set

        self._session: rio.Session | None = None
        self._transport = MessageRecorderTransport(
            process_sent_message=self._process_sent_message
        )
        self._first_refresh_completed = asyncio.Event()

    def _process_sent_message(self, message: JsonDoc) -> None:
        if "id" in message:
            self._transport.queue_response(
                {
                    "jsonrpc": "2.0",
                    "id": message["id"],
                    "result": None,
                }
            )

        if message["method"] == "updateComponentStates":
            self._first_refresh_completed.set()

    async def __aenter__(self) -> te.Self:
        url = str(rio.URL("http://unit.test") / self._active_url.lstrip("/"))

        self._session = await self._app_server.create_session(
            initial_message=data_models.InitialClientMessage.from_defaults(
                url=url,
                user_settings=self._user_settings,
            ),
            transport=self._transport,
            client_ip="localhost",
            client_port=12345,
            http_headers=starlette.datastructures.Headers(),
        )

        if self._use_ordered_dirty_set:
            self._session._dirty_components = ordered_set.OrderedSet(
                self._session._dirty_components
            )  # type: ignore

        await self._first_refresh_completed.wait()

        return self

    async def __aexit__(self, *_) -> None:
        if self._session is not None:
            await self._session._close(close_remote_session=False)

    async def _simulate_interrupted_connection(self) -> None:
        assert self._session is not None

        self._transport.queue_response(TransportInterrupted)

        while self._session._is_connected_event.is_set():
            await asyncio.sleep(0.05)

    async def _simulate_reconnect(self) -> None:
        assert self._session is not None

        # If currently connected, disconnect first
        if self._session._is_connected_event.is_set():
            await self._simulate_interrupted_connection()

        self._transport = MessageRecorderTransport(
            process_sent_message=self._process_sent_message
        )
        self._session._transport = self._transport

    @property
    def _outgoing_messages(self) -> list[JsonDoc]:
        return self._transport.sent_messages

    @property
    def _dirty_components(self) -> set[rio.Component]:
        return set(self.session._dirty_components)

    @property
    def _last_updated_components(self) -> set[rio.Component]:
        return set(self._last_component_state_changes)

    @property
    def _last_component_state_changes(
        self,
    ) -> t.Mapping[rio.Component, t.Mapping[str, object]]:
        for message in reversed(self._transport.sent_messages):
            if message["method"] == "updateComponentStates":
                delta_states: dict = message["params"]["deltaStates"]  # type: ignore
                return {
                    self.session._weak_components_by_id[
                        int(component_id)
                    ]: delta
                    for component_id, delta in delta_states.items()
                    if int(component_id)
                    != self.session._high_level_root_component._id
                }

        return {}

    def _get_build_output(
        self,
        component: rio.Component,
        type_: type[C] | None = None,
    ) -> C:
        result = component._build_data_.build_result  # type: ignore

        if type_ is not None:
            assert (
                type(result) is type_
            ), f"Expected {type_}, got {type(result)}"

        return result  # type: ignore

    @property
    def session(self) -> rio.Session:
        assert self._session is not None

        return self._session

    @property
    def crashed_build_functions(self) -> t.Mapping[t.Callable, str]:
        return self.session._crashed_build_functions

    @property
    def root_component(self) -> rio.Component:
        return self.session._get_user_root_component()

    def get_components(self, component_type: type[C]) -> t.Iterator[C]:
        root_component = self.root_component

        for component in root_component._iter_component_tree_():
            if type(component) is component_type:
                yield component  # type: ignore

    def get_component(self, component_type: type[C]) -> C:
        try:
            return next(self.get_components(component_type))
        except StopIteration:
            raise AssertionError(f"No component of type {component_type} found")

    async def refresh(self) -> None:
        await self.session._refresh()



import typing as t

import pytest

import rio
from tests.utils.layouting import verify_layout


@pytest.mark.parametrize(
    "container_type",
    [rio.Row, rio.Column],
)
async def test_linear_container_with_no_extra_width(
    container_type: type,
) -> None:
    await verify_layout(
        lambda: container_type(
            rio.Text("hi", min_width=100),
            rio.Button("clicky", min_width=400),
        )
    )


@pytest.mark.parametrize(
    "horizontal",
    [True, False],
)
@pytest.mark.parametrize(
    "first_child_grows",
    [True, False],
)
@pytest.mark.parametrize(
    "second_child_grows",
    [True, False],
)
@pytest.mark.parametrize(
    "proportions",
    [
        None,
        "homogeneous",
        [1, 2],
        [2, 1],
    ],
)
async def test_linear_container_with_extra_width(
    horizontal: bool,
    first_child_grows: bool,
    second_child_grows: bool,
    proportions: None | t.Literal["homogeneous"] | list[int],
) -> None:
    """
    A battery of scenarios to test the most common containers - Rows & Columns.
    """
    if horizontal:
        container_type = rio.Row

        if first_child_grows:
            first_child_width = 0
            first_child_grow_x = True
        else:
            first_child_width = 10
            first_child_grow_x = False

        if second_child_grows:
            second_child_width = 0
            second_child_grow_x = True
        else:
            second_child_width = 20
            second_child_grow_x = False

        first_child_height = 0
        second_child_height = 0

        first_child_grow_y = False
        second_child_grow_y = False

        parent_width = 50
        parent_height = 0
    else:
        container_type = rio.Column

        if first_child_grows:
            first_child_height = 0
            first_child_grow_y = True
        else:
            first_child_height = 10
            first_child_grow_y = False

        if second_child_grows:
            second_child_height = 0
            second_child_grow_y = True
        else:
            second_child_height = 20
            second_child_grow_y = False

        first_child_width = 0
        second_child_width = 0

        first_child_grow_x = False
        second_child_grow_x = False

        parent_width = 0
        parent_height = 50

    await verify_layout(
        lambda: container_type(
            rio.Text(
                "short-text",
                min_width=first_child_width,
                min_height=first_child_height,
                grow_x=first_child_grow_x,
                grow_y=first_child_grow_y,
            ),
            rio.Text(
                "very-much-longer-text",
                min_width=second_child_width,
                min_height=second_child_height,
                grow_x=second_child_grow_x,
                grow_y=second_child_grow_y,
            ),
            # It would be nice to vary the spacing as well, but that would once
            # again double the number of tests this case already has. Simply
            # always specify a spacing, since that is the harder case anyway.
            spacing=2,
            proportions=proportions,
            min_width=parent_width,
            min_height=parent_height,
            align_x=0.5,
            align_y=0.5,
        )
    )



"""
Project templates are stored using Rio's snippet system.

This file ensures that the snippets for project templates match expectations.
"""

from __future__ import annotations

import tempfile
import typing as t
from pathlib import Path

import pytest

import rio.cli
import rio.snippets


def test_empty_template_exists() -> None:
    """
    The "Empty" template is special, because it is used as a fallback if no
    real template is selected. It must exist.
    """

    for template in rio.snippets.get_project_templates(include_empty=True):
        if template.name == "Empty":
            return

    raise AssertionError("The `Empty` template is missing.")


def test_template_structure_is_valid() -> None:
    """
    Ensures that the snippets for project templates match expectations, by
    parsing them.
    """
    # Iterate over all templates. This forces them to be parsed
    for _ in rio.snippets.get_project_templates(include_empty=True):
        pass


def test_available_template_literal_matches_templates() -> None:
    """
    Rio defines a literal type which corresponds to all available project
    templates. This test verifies that the literal and the actual templates
    match.
    """
    # Find all templates according to the literal
    templates_according_to_literal = set(
        t.get_args(rio.snippets.AvailableTemplatesLiteral)
    ) | {"Empty"}

    # Find all defined templates
    templates_according_to_snippets = {
        template.name
        for template in rio.snippets.get_project_templates(include_empty=True)
    }

    # They must match
    assert templates_according_to_literal == templates_according_to_snippets


@pytest.mark.parametrize(
    "template",
    rio.snippets.get_project_templates(include_empty=True),
)
def test_instantiate_template(template: rio.snippets.ProjectTemplate) -> None:
    """
    Instantiates all templates to ensure that they can be instantiated without
    crashing.
    """

    # Create a temporary directory for the project
    with tempfile.TemporaryDirectory() as project_directory_str:
        # Instantiate the template
        rio.cli.project_setup.create_project(
            raw_name=f"Test Project {template.name}",
            type="website",
            template_name=template.name,
            target_parent_directory=Path(project_directory_str),
        )

        # There are no further checks here. The test simply makes sure there is
        # no crash during the instantiation.



from __future__ import annotations

import dataclasses
import typing as t

from uniserde import Jsonable

import rio

from . import color, deprecations
from . import text_style as text_style_module

__all__ = [
    "Palette",
    "Theme",
]


T = t.TypeVar("T")


def _derive_color(
    color: rio.Color,
    offset: float,
    *,
    bias_to_bright: float = 0,
    target_color: rio.Color | None = None,
) -> rio.Color:
    """
    Given a base color, derive a related, but different color. The idea is to
    shift the color slightly to create a new color that can be used to create
    visual interest.

    ## Params

    `base_color`: The color to derive a new color from.

    `offset`: Controls how much the color should be shifted. A value of `0` will
        return the original color, while a value of `1` will return a massively
        different color.

    `bias_to_bright`: A value between `-1` and `1` that controls how much the
        new color should be biased towards brighter or darker colors. A value
        of `-1` will always return a darker color, while a value of `1` will
        always return a brighter color.

    `target_color`: If provided, the new color will be biased towards this
        color. This can be used to tint the color towards a specific hue.
    """
    assert 0 <= offset <= 1, f"The offset must be between 0 and 1, not {offset}"
    assert (
        -1 <= bias_to_bright <= 1
    ), f'The "bias_to_bright" must be between -1 and 1, not {bias_to_bright}'

    # If a target color was provided, move towards that color
    #
    # The more different the two colors are, the less the new color will be
    # used. This helps e.g. with dark themes, which look bonkers if the target
    # color is very light.
    if target_color is not None:
        difference = abs(
            target_color.perceived_brightness - color.perceived_brightness
        )

        if difference < 0.01:
            offset_scale = 1
        else:
            offset_scale = min(1.5 / difference, 1)

        result = color.blend(target_color, offset * offset_scale)

    # Otherwise change the color's brightness
    else:
        threshold = 0.5 + 0.5 * bias_to_bright
        perceived_brightness = color.perceived_brightness
        brighten = perceived_brightness <= threshold

        # Calculate the new color
        if brighten:
            result = color.brighter(offset)
        else:
            result = color.darker(offset)

    return result


def _make_semantic_palette(color: rio.Color) -> Palette:
    return Palette(
        background=color,
        background_variant=_derive_color(
            color,
            0.1,
            bias_to_bright=-0.4,
        ),
        background_active=_derive_color(
            color,
            0.2,
            bias_to_bright=0.8,
        ),
        foreground=_derive_color(
            color,
            0.6,
        ),
    )


@t.final
@dataclasses.dataclass()
class Palette:
    """
    A group of related colors in a theme.

    Rio's themes consist of multiple palettes. Each palette has the same set of
    colors, but in different shades. You can switch between palettes using a
    `rio.ThemeContextSwitcher`.

    Warning: Future versions of Rio may add additional colors to palettes. For
        this reason, there is no constructor that simply accepts all palette
        colors. This way any missing colors can be automatically derived.

    ## Attributes

    `background`: The main background color of the palette. This color is used
        for wide sweeping regions of the app, such as cards or the background of
        the app.

        In semantic palettes ("Success", "Warning", "Danger"), this color is
        also the color you're probably looking for. For example, if you want to
        indicate danger, the danger palette's background color is the one to
        use.

    `background_variant`: Similar to the background color, but slightly
        modified. This is used by many components to fit the background while
        still providing enough contrast. For example, this is the main color of
        a `rio.TextInput`.

    `background_active`: Used by components to indicate interactivity. Fora
        example a `rio.TextInput` might switch to this when hovered.

    `foreground`: The color to use for text and other elements that need to
        be legible or otherwise stand out. This is what you'll typically use for
        text and icons.
    """

    background: rio.Color
    background_variant: rio.Color
    background_active: rio.Color

    foreground: rio.Color

    def __init__(
        self,
        background: rio.Color,
        *,
        background_variant: rio.Color | None = None,
        background_active: rio.Color | None = None,
        foreground: rio.Color | None = None,
        contrast: float = 1,
        target_color: rio.Color | None = None,
    ) -> None:
        """
        Creates a new palette from the provided color.

        Creates a new palette, deriving any missing colors from the provided
        background color.

        By default, the palette colors will be similar to the background color,
        but with slightly different shades. If you provide a `target_color`, the
        colors will instead be derived by blending in that color.

        ## Parameters

        `background`: The main background color of the palette. All other colors
            will be derived assuming components using this palette are placed on
            a large area of this color.

        `background_variant`: A slightly modified version of the background
            color. This is used by many components to fit the background while
            still providing enough contrast. For example, this is the main color
            of a `rio.TextInput`. This will be derived automatically if not
            provided.

        `background_active`: Used by components to indicate interactivity. For
            example a `rio.TextInput` might switch to this when hovered. This
            will be derived automatically if not provided.

        `foreground`: The color to use for text and other elements that need to
            be legible or otherwise stand out. This is what you'll typically use
            for text and icons. This will be derived automatically if not
            provided.

        `contrast`: When deriving colors, this controls how much the colors
            should differ from the background. Values less than `1` will create
            less contrast than would be the default, while values greater than
            `1` will create more contrast.

        `target_color`: If not provided, derived colors will be different shades
            of the background color. If provided, the colors will be derived by
            blending in this color. This can be used to tint the colors towards
            a specific hue.
        """

        self.background = background

        if background_variant is None:
            self.background_variant = _derive_color(
                background,
                offset=0.1 * contrast,
                bias_to_bright=-0.3,
                target_color=target_color,
            )
        else:
            self.background_variant = background_variant

        if background_active is None:
            self.background_active = _derive_color(
                background,
                offset=0.2 * contrast,
                bias_to_bright=-0.3,
                target_color=target_color,
            )
        else:
            self.background_active = background_active

        if foreground is None:
            self.foreground = _derive_color(
                background,
                offset=0.8,
            )
        else:
            self.foreground = foreground

    @staticmethod
    @deprecations.deprecated(
        since="0.11.0",
        description="The `Palette` constructor now has the functionality that `from_color` used to provide. Pass your color to the constructor directly.",
    )
    def from_color(
        color: rio.Color,
        *,
        offset: float = 1,
        target_color: rio.Color | None = None,
    ) -> Palette:
        """
        Derives a palette from a single color.

        This creates a new palette from the provided color. Any colors needed by
        palettes that aren't specified are automatically derived to fit in.

        # Parameters

        `color`: The main background color of the palette. All other colors
            will be derived assuming components using this palette are placed on
            a large area of this color.

        `offset`: When deriving colors, this controls how much the colors should
            differ from the background. Values less than `1` will create less
            contrast than would be the default, while values greater than `1`
            will create more contrast

        `target_color`: If not provided, derived colors will be different shades
            of the background color. If provided, the colors will be derived by
            blending in this color. This can be used to tint the colors towards
            a specific hue.
        """
        return Palette(
            background=color,
            contrast=offset,
            target_color=target_color,
        )


@t.final
@dataclasses.dataclass()
class Theme:
    """
    Defines the visual style of the application.

    The `Theme` contains all colors, text styles, and other visual properties
    that are used throughout the application. If you wish to change the
    appearance of your app, this is the place to do it.

    Warning: The exact attributes available in themes are still subject to
        change. The recommended way to create themes is using either the
        `from_colors` or `pair_from_colors` method, as they provide a more
        stable interface.

    ## Attributes

    `primary_palette`: The main (non-neutral) color of your app. This is
        sprinkled in throughout the theme to create a cohesive look.

    `secondary_palette`: A color that complements the primary color. This is
        often used for smaller components such as buttons and switches.

    `background_palette`: A neutral color that will be used to fill large areas
        of empty space. This is also the app's background color.

    `neutral_palette`: Similar to the background color, this is also used to
        fill large areas areas, that need to be distinctive from the background.
        For example, `rio.Card` uses this color by default.

    `hud_palette`: Used for elements that pop over the content, such as
        tooltips.

    `disabled_palette`: Used by insensitive components to indicate that they are
        not interactive. Typically consists of shades of gray.

    `success_palette`: A color to give positive feedback the user. Typically
        consists of shades of green.

    `warning_palette`: A color to indicate that something might be wrong, but
        isn't critical. Typically consists of shades of orange.

    `danger_palette`: A color to indicate that something is wrong and needs
        immediate attention. Typically consists of shades of red.

    `corner_radius_small`: The corner radius of small components such as text
        inputs.

    `corner_radius_medium`: The corner radius of medium-sized components, such
        as small cards.

    `corner_radius_large`: The corner radius of large components, such as large
        cards and dialogs.

    `shadow_color`: The color to use for shadows. This is typically a dark
        color.

    `monospace_font`: The font to use for monospace text, such as code.

    `heading1_style`: The text style to use for the largest headings.

    `heading2_style`: The text style to use for the second largest headings.

    `heading3_style`: The text style to use for the third largest headings.

    `text_style`: The text style to use for regular text.
    """

    _: dataclasses.KW_ONLY

    primary_palette: Palette
    secondary_palette: Palette

    background_palette: Palette
    neutral_palette: Palette
    hud_palette: Palette
    disabled_palette: Palette

    success_palette: Palette
    warning_palette: Palette
    danger_palette: Palette

    # Other
    corner_radius_small: float
    corner_radius_medium: float
    corner_radius_large: float

    shadow_color: rio.Color

    monospace_font: text_style_module.Font

    # Text styles
    heading1_style: rio.TextStyle
    heading2_style: rio.TextStyle
    heading3_style: rio.TextStyle
    text_style: rio.TextStyle

    def __init__(self) -> None:
        # Themes are still very much in flux. New attributes will be added,
        # removed and changed. Stop anyone from trying to create a theme
        # manually, as that is an endeavor doomed to fail.
        """
        Do not use. Themes are still in flux and should be created using the
        `from_colors` or `pair_from_colors` methods.

        ## Metadata

        `public`: False
        """

    @staticmethod
    def _create_new(
        primary_palette: Palette,
        secondary_palette: Palette,
        background_palette: Palette,
        neutral_palette: Palette,
        hud_palette: Palette,
        disabled_palette: Palette,
        success_palette: Palette,
        warning_palette: Palette,
        danger_palette: Palette,
        corner_radius_small: float,
        corner_radius_medium: float,
        corner_radius_large: float,
        shadow_color: rio.Color,
        monospace_font: text_style_module.Font,
        heading1_style: rio.TextStyle,
        heading2_style: rio.TextStyle,
        heading3_style: rio.TextStyle,
        text_style: rio.TextStyle,
    ) -> Theme:
        """
        The Theme's constructor is blocked to prevent users from creating themes
        manually. This is an internal method that bypasses that block.
        """
        self = object.__new__(Theme)

        self.__dict__.update(
            {
                "primary_palette": primary_palette,
                "secondary_palette": secondary_palette,
                "background_palette": background_palette,
                "neutral_palette": neutral_palette,
                "hud_palette": hud_palette,
                "disabled_palette": disabled_palette,
                "success_palette": success_palette,
                "warning_palette": warning_palette,
                "danger_palette": danger_palette,
                "corner_radius_small": corner_radius_small,
                "corner_radius_medium": corner_radius_medium,
                "corner_radius_large": corner_radius_large,
                "shadow_color": shadow_color,
                "monospace_font": monospace_font,
                "heading1_style": heading1_style,
                "heading2_style": heading2_style,
                "heading3_style": heading3_style,
                "text_style": text_style,
            }
        )

        return self

    @classmethod
    @deprecations.parameter_remapped(
        since="0.9.0",
        old_name="light",
        new_name="mode",
        remap=lambda light: "light" if light else "dark",
    )
    def from_colors(
        cls,
        *,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.4,
        corner_radius_medium: float = 0.8,
        corner_radius_large: float = 1.8,
        heading_fill: t.Literal["primary", "plain", "auto"]
        | text_style_module._TextFill = "auto",
        text_color: rio.Color | None = None,
        font: rio.Font = text_style_module.Font.ROBOTO,
        heading_font: rio.Font | None = None,
        monospace_font: text_style_module.Font = text_style_module.Font.ROBOTO_MONO,
        mode: t.Literal["light", "dark"] = "light",
    ) -> rio.Theme:
        """
        Creates a new theme based on the provided colors.

        Themes store a large number of colors and text styles that are used
        throughout the app. This function provides a convenient way to only
        specify the most important colors of your theme, and have Rio handle the
        rest. It is the recommended way to create a new theme.

        To use a custom theme in your app first create a theme, and then pass it
        to your app instance

        ```python
        # Create a new theme
        theme = rio.Theme.from_colors(
            # Configure your theme here
        )

        # And apply it to your app
        app = rio.App(
            theme=theme,
            # ...
        )
        ```

        Alternatively, there is also `pair_from_colors`, which creates both a
        light and a dark theme at the same time. This is useful if you want Rio
        to automatically switch between the two based on the user's system
        preferences.


        ## Parameters

        `primary_color`: The main color of your app. This color will be used to
            tint the background and by some large components to fill large
            spaces with color.

        `secondary_color`: A color that nicely complements the primary color. It
            is often used by small components such as buttons and switches.

        `background_color`: The app's background color. This should be a neutral
            color that doesn't distract from the content.

        `neutral_color`: Similar to the background color, it is also used for
            neutral areas. It should however be slightly different, allowing you
            to create a visual hierarchy. This is the default color of large
            elements such as cards.

        `hud_color`: Used for elements that pop over the content, such as
            tooltips.

        `disabled_color`: Used by insensitive components to indicate that they
            are not interactive. Typically a shade of gray.

        `success_color`: A color to give positive feedback the user. Typically
            a shade of green.

        `warning_color`: A color to indicate that something might be wrong, but
            isn't critical. Typically orange.

        `danger_color`: A color to indicate that something is wrong and needs
            immediate attention. Typically a shade of red.

        `corner_radius_small`: The corner radius of small components such as
            text inputs

        `corner_radius_medium`: The corner radius of medium-sized components,
            such as small cards.

        `corner_radius_large`: The corner radius of large components, such as
            large cards and dialogs.

        `heading_fill`: The fill to use for headings. This allows you to specify
            a more interesting color, or even a gradient. If set to `"auto"`,
            Rio will automatically switch between the primary color and a plain
            text color based on legibility.

            This only affects headings in background and neutral contexts.

        `text_color`: The default text color to use for regular text. Please
            note that this only applies to text in a neutral or background
            context. Text that's e.g. placed on a `rio.Card` with
            `color="primary"` will use a different color to ensure legibility.

        `font`: The default font to use when no other is specified.

        `heading_font`: The font to use for text headings ("heading1",
            "heading2", ...). If not provided, the regular font will be used

        `monospace_font`: The font to use for monospace text, such as code.

        `mode`: Whether to create a light or dark theme. This affects the
            default values for some colors, such as the background.
        """

        # Primary palette
        if primary_color is None:
            primary_color = rio.Color.from_hex("01dffd")

        primary_palette = Palette(primary_color)

        # Secondary palette
        if secondary_color is None:
            secondary_color = rio.Color.from_hex("0083ff")

        secondary_palette = Palette(secondary_color)

        # Background palette
        if background_color is None:
            if mode == "light":
                background_color = rio.Color.from_rgb(0.96, 0.96, 0.93)
            else:
                background_color = rio.Color.from_gray(0.10)

        if text_color is None:
            neutral_and_background_text_color = (
                # Gray tones look good on bright themes
                rio.Color.from_gray(0.3)
                if background_color.perceived_brightness > 0.5
                # ... but not on dark ones. Go very bright here.
                else rio.Color.from_gray(0.9)
            )
        else:
            neutral_and_background_text_color = text_color

        del text_color

        if neutral_color is None:
            neutral_color = _derive_color(
                background_color,
                0.04,
            ).blend(primary_color, 0.06)

        background_palette = Palette(
            background=background_color,
            background_variant=neutral_color,
            background_active=_derive_color(
                background_color,
                0.25,
                bias_to_bright=0.15,
                target_color=primary_color,
            ),
            foreground=neutral_and_background_text_color,
        )

        # Neutral palette
        neutral_palette = Palette(
            background=neutral_color,
            background_variant=_derive_color(
                neutral_color,
                0.15,
                bias_to_bright=0.15,
                target_color=primary_color,
            ),
            background_active=_derive_color(
                neutral_color,
                0.25,
                bias_to_bright=0.15,
                target_color=primary_color,
            ),
            foreground=neutral_and_background_text_color,
        )

        # HUD palette
        if hud_color is None:
            if mode == "light":
                hud_color = rio.Color.from_gray(0.15)
            else:
                hud_color = rio.Color.from_gray(0.02)

        hud_palette = Palette(
            background=hud_color,
            background_variant=_derive_color(
                hud_color,
                0.08,
            ),
            background_active=_derive_color(
                hud_color,
                0.15,
            ),
            foreground=(
                rio.Color.from_gray(0.1)
                if hud_color.perceived_brightness > 0.5
                else rio.Color.from_gray(0.9)
            ),
        )

        # Keep the disabled palette subdued. It's not meant to be perfectly
        # readable
        if disabled_color is None:
            disabled_color = (
                neutral_palette.background.blend(
                    neutral_palette.foreground, 0.35
                )
            ).desaturated(0.3)

        disabled_palette = Palette(
            background=disabled_color,
            background_variant=disabled_color.blend(
                neutral_palette.background, 0.35
            ),
            background_active=disabled_color.blend(
                neutral_palette.background, 0.5
            ),
            foreground=_derive_color(disabled_color, 0.3),
        )

        # Shadow color
        if mode == "light":
            shadow_color = rio.Color.from_rgb(0.1, 0.1, 0.2, 0.5)
        else:
            shadow_color = rio.Color.BLACK

        # Semantic colors
        if success_color is None:
            success_color = rio.Color.from_hex("1e8e3e")

        if warning_color is None:
            warning_color = rio.Color.from_hex("f9a825")

        if danger_color is None:
            danger_color = rio.Color.from_hex("b3261e")

        success_palette = _make_semantic_palette(success_color)
        warning_palette = _make_semantic_palette(warning_color)
        danger_palette = _make_semantic_palette(danger_color)

        # Colorful headings can be a problem when the primary color is similar
        # to the background/neutral color. If the `color_headings` argument is
        # set to `auto`, disable coloring if the colors are close.
        if heading_fill == "auto":
            brightness1 = primary_palette.background.perceived_brightness
            brightness2 = background_palette.background.perceived_brightness

            heading_fill = (
                "primary" if abs(brightness1 - brightness2) > 0.3 else "plain"
            )

        if heading_fill == "primary":
            heading_fill = primary_color
        elif heading_fill == "plain":
            heading_fill = neutral_and_background_text_color
        else:
            heading_fill = heading_fill

        # Text styles
        heading1_style = rio.TextStyle(
            font=font if heading_font is None else heading_font,
            fill=heading_fill,
            font_size=2.3,
        )
        heading2_style = heading1_style.replace(font_size=1.7)
        heading3_style = heading1_style.replace(font_size=1.2)
        text_style = heading1_style.replace(
            font=font,
            font_size=1,
            fill=neutral_and_background_text_color,
        )

        # Build the final theme
        # Instantiate the theme. `__init__` is blocked to prevent users from
        # doing something foolish. Work around that.
        return rio.Theme._create_new(
            primary_palette=primary_palette,
            secondary_palette=secondary_palette,
            background_palette=background_palette,
            neutral_palette=neutral_palette,
            hud_palette=hud_palette,
            disabled_palette=disabled_palette,
            success_palette=success_palette,
            warning_palette=warning_palette,
            danger_palette=danger_palette,
            corner_radius_small=corner_radius_small,
            corner_radius_medium=corner_radius_medium,
            corner_radius_large=corner_radius_large,
            shadow_color=shadow_color,
            monospace_font=monospace_font,
            heading1_style=heading1_style,
            heading2_style=heading2_style,
            heading3_style=heading3_style,
            text_style=text_style,
        )

    @classmethod
    def pair_from_colors(
        cls,
        *,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.4,
        corner_radius_medium: float = 0.8,
        corner_radius_large: float = 1.8,
        text_color: rio.Color
        | tuple[rio.Color | None, rio.Color | None]
        | None = None,
        font: text_style_module.Font = text_style_module.Font.ROBOTO,
        heading_font: rio.Font | None = None,
        monospace_font: text_style_module.Font = text_style_module.Font.ROBOTO_MONO,
        heading_fill: t.Literal["primary", "plain", "auto"]
        | text_style_module._TextFill = "auto",
    ) -> tuple[Theme, Theme]:
        """
        This function is very similar to `from_colors`, but it returns two
        themes: A light and a dark one. When applying two themes to your app,
        Rio will automatically switch between them based on the user's system
        preferences.

        ```python
        # Create a theme pair
        themes = rio.Theme.pair_from_colors(
            # Configure your theme here
        )

        # And apply them to your app
        app = rio.App(
            theme=themes,
            # ...
        )
        ```

        ## Parameters

        `primary_color`: The main color of your app. This color will be used to
            tint the background and by some large components to fill large
            spaces with color

        `secondary_color`: A color that nicely complements the primary color. It
            is often used by small components such as buttons and switches.

        `background_color`: The app's background color. This should be a neutral
            color that doesn't distract from the content.

        `neutral_color`: Similar to the background color, it is also used for
            neutral areas. It should however be slightly different, allowing you
            to create a visual hierarchy. This is the default color of large
            elements such as cards.

        `hud_color`: Used for elements that pop over the content, such as
            tooltips.

        `disabled_color`: Used by insensitive components to indicate that they
            are not interactive. Typically a shade of gray.

        `success_color`: A color to give positive feedback the user. Typically
            a shade of green.

        `warning_color`: A color to indicate that something might be wrong, but
            isn't critical. Typically orange.

        `danger_color`: A color to indicate that something is wrong and needs
            immediate attention. Typically a shade of red.

        `corner_radius_small`: The corner radius of small components such as
            text inputs

        `corner_radius_medium`: The corner radius of medium-sized components,
            such as small cards.

        `corner_radius_large`: The corner radius of large components, such as
            large cards and dialogs.

        `heading_fill`: The fill to use for headings. This allows you to specify
            a more interesting color, or even a gradient. If set to `"auto"`,
            Rio will automatically switch between the primary color and a plain
            text color based on legibility.

            This only affects headings in background and neutral contexts.

        `text_color`: The default text color to use for regular text. Please
            note that this only applies to text in a neutral or background
            context. Text that's e.g. placed on a `rio.Card` with
            `color="primary"` will use a different color to ensure legibility.

            You can also specify a tuple of two colors to use different text
            colors for light and dark themes.

        `font`: The default font to use when no other is specified.

        `heading_font`: The font to use for text headings ("heading1",
            "heading2", ...). If not provided, the regular font will be used

        `monospace_font`: The font to use for monospace text, such as code.
        """
        if not isinstance(text_color, tuple):
            text_color = (text_color, text_color)

        themes = list[Theme]()
        for mode, text_color in zip(("light", "dark"), text_color):
            themes.append(
                cls.from_colors(
                    primary_color=primary_color,
                    secondary_color=secondary_color,
                    background_color=background_color,
                    neutral_color=neutral_color,
                    hud_color=hud_color,
                    disabled_color=disabled_color,
                    success_color=success_color,
                    warning_color=warning_color,
                    danger_color=danger_color,
                    corner_radius_small=corner_radius_small,
                    corner_radius_medium=corner_radius_medium,
                    corner_radius_large=corner_radius_large,
                    font=font,
                    heading_font=heading_font,
                    monospace_font=monospace_font,
                    heading_fill=heading_fill,
                    text_color=text_color,
                    mode=mode,
                )
            )

        return tuple(themes)  # type: ignore (length mismatch)

    def text_color_for(self, color: rio.Color) -> rio.Color:
        """
        Derive a color legible on the given one.

        Given the color of a background, return a legible text color to use on
        top of it.

        ## Parameters

        `color`: The color of the background on which the text should be
            legible.
        """
        return _derive_color(color, offset=0.8)

    def _serialize_colorset(self, color: color.ColorSet) -> Jsonable:
        # A colorset more-or-less translates to a switcheroo. Thus, this
        # function needs to provide all information needed to create one.

        # If the color is a string, just pass it through. In this case there is
        # an actual switcheroo which can be used.
        if isinstance(color, str):
            return color

        # TODO: If the color is identical/very similar to a theme color use the
        #       corresponding switcheroo.

        # Otherwise create all the necessary variables to emulate a switcheroo.
        palette = Palette(color)

        return {
            "localBg": palette.background.srgba,
            "localBgVariant": palette.background_variant.srgba,
            "localBgActive": palette.background_active.srgba,
            "localFg": palette.foreground.srgba,
        }

    @property
    def is_light_theme(self) -> bool:
        """
        Whether the theme is a light theme.

        Returns `True` if the theme is a light theme, and `False` otherwise.
        This isn't stored in the theme itself, but is instead derived from the
        background color. If the background color is bright, this property will
        return `True`.
        """
        return self.background_palette.background.perceived_brightness >= 0.5

    @property
    def primary_color(self) -> rio.Color:
        """
        The most important accent color of the theme.

        Returns the configured primary color of the theme. This is the same as
        `theme.primary_palette.background`.
        """
        return self.primary_palette.background

    @property
    def secondary_color(self) -> rio.Color:
        """
        A complementary accent color of the theme.

        Returns the configured secondary color of the theme. This is the same as
        `theme.secondary_palette.background`.
        """
        return self.secondary_palette.background

    @property
    def background_color(self) -> rio.Color:
        """
        A color to fill large areas of empty space.

        Returns the configured background color of the theme. This is the same
        as `theme.background_palette.background`.
        """
        return self.background_palette.background

    @property
    def neutral_color(self) -> rio.Color:
        """
        Another color to fill large areas of empty space.

        Returns the configured neutral color of the theme. This is the same as
        `theme.neutral_palette.background`.
        """
        return self.neutral_palette.background

    @property
    def hud_color(self) -> rio.Color:
        """
        A color to use for elements that pop over the content.

        Returns the configured HUD color of the theme. This is the same as
        `theme.hud_palette.background`.
        """
        return self.hud_palette.background

    @property
    def disabled_color(self) -> rio.Color:
        """
        A color to use for elements that are not interactive.

        Returns the configured disabled color of the theme. This is the same as
        `theme.disabled_palette.background`.
        """
        return self.disabled_palette.background

    @property
    def success_color(self) -> rio.Color:
        """
        A color to give positive feedback to the user.

        Returns the configured success color of the theme. This is the same as
        `theme.success_palette.background`.
        """
        return self.success_palette.background

    @property
    def warning_color(self) -> rio.Color:
        """
        A color to indicate that something might be wrong.

        Returns the configured warning color of the theme. This is the same as
        `theme.warning_palette.background`.
        """
        return self.warning_palette.background

    @property
    def danger_color(self) -> rio.Color:
        """
        A color to indicate that something is wrong and needs attention.

        Returns the configured danger color of the theme. This is the same as
        `theme.danger_palette.background`.
        """
        return self.danger_palette.background

    @property
    def font(self) -> text_style_module.Font:
        """
        The font to use for regular text.

        Returns the configured default font of the theme. This is used for
        most text in the app, though headers, code and other special cases may
        choose to use different fonts.
        """
        assert (
            self.text_style.font is not None
        ), f"The theme's text style must have a font set"
        return self.text_style.font



from __future__ import annotations

import typing as t

import typing_extensions as te

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["Stack"]


@t.final
class Stack(FundamentalComponent):
    """
    A container that stacks its children in the Z direction.

    `Stacks` are similar to rows and columns, but they stack their children in
    the Z direction instead of the X or Y direction. In other words, the stack's
    children overlap each other, with the first one at the bottom, the second
    one above that, and so on.


    ## Attributes

    `children`: The components to place in this `Stack`.


    ## Examples

    This example will create a stack of three icons, each with a different size
    and color. The first icon will be at the bottom, the second one above that,
    and the third one at the top:

    ```python
    rio.Stack(
        # bottom
        rio.Icon(
            "material/castle",
            min_width=50,
            min_height=50,
            fill=rio.Color.from_hex("00ff00"),
            align_x=0.5,
            align_y=0.5,
        ),
        # middle
        rio.Icon(
            "material/castle",
            min_width=30,
            min_height=30,
            fill=rio.Color.from_hex("ff0000"),
            align_x=0.5,
            align_y=0.5,
        ),
        # top
        rio.Icon(
            "material/castle",
            min_width=10,
            min_height=10,
            fill=rio.Color.from_hex("000000"),
            align_x=0.5,
            align_y=0.5,
        ),
    )
    ```
    """

    children: list[rio.Component]

    def __init__(
        self,
        *children: rio.Component,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.children = list(children)

    def add(self, child: rio.Component) -> te.Self:
        """
        Appends a child component.

        Appends a child component to the end and then returns the `Stack`, which
        makes method chaining possible:

        ```python
        rio.Stack().add(child1).add(child2)
        ```

        ## Parameters

        `child`: The child component to append.
        """
        self.children.append(child)
        return self


Stack._unique_id_ = "Stack-builtin"



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
class Footer(rio.Component):
    """
    A responsive footer component that displays company-related information and
    navigation links.

    The footer is organized into sections for resources, features, company
    information, and a newsletter subscription option. The layout adapts to
    desktop and mobile devices.
    """

    def build(self) -> rio.Component:
        """
        Constructs the footer layout based on the device type.

        For desktop:
        - Displays sections in a row layout with a minimum width.

        For mobile:
        - Adjusts the layout to a column with additional margins and spacing.
        """
        # Get the device type from the session
        device = self.session[data_models.PageLayout].device

        # Define the main content layout based on the device type
        if device == "desktop":
            content = rio.Row(
                margin_top=2,
                align_x=0.5,
                min_width=80,
            )
        else:
            content = rio.Row(
                spacing=2,
            )

        # Add the Resources section
        content.add(
            rio.Column(
                rio.Text(
                    "Ressources",
                    style=rio.TextStyle(font_weight="bold", font_size=0.9),
                    margin_bottom=0.5,
                ),
                comps.FakeLink("Help center"),
                comps.FakeLink("Docs"),
                comps.FakeLink("Roadmap"),
                comps.FakeLink("Changelog"),
                spacing=1,
            ),
        )

        # Add the Features section
        content.add(
            rio.Column(
                rio.Text(
                    "Features",
                    style=rio.TextStyle(font_weight="bold", font_size=0.9),
                    margin_bottom=1,
                ),
                comps.FakeLink("Affiliates"),
                comps.FakeLink("Portal"),
                comps.FakeLink("Jobs"),
                comps.FakeLink("Sponsors"),
                spacing=1,
            ),
        )

        # Add the Company section
        content.add(
            rio.Column(
                rio.Text(
                    "Company",
                    style=rio.TextStyle(font_weight="bold", font_size=0.9),
                    margin_bottom=1,
                ),
                comps.FakeLink("About"),
                comps.FakeLink("Pricing"),
                comps.FakeLink("Careers"),
                comps.FakeLink("Blog"),
                spacing=1,
            ),
        )

        # Add the newsletter subscription section
        if device == "desktop":
            content.add(
                rio.Column(
                    rio.Text(
                        "Subscribe to our newsletter",
                        style=rio.TextStyle(font_weight="bold"),
                    ),
                    comps.SubscribeInputButton(),
                    spacing=1,
                    align_y=0,
                ),
            )
        else:
            # Adjust the layout for mobile with the subscription section
            content = rio.Column(
                content,  # Existing content
                rio.Column(
                    rio.Text(
                        "Subscribe to our newsletter",
                        style=rio.TextStyle(font_weight="bold"),
                    ),
                    comps.SubscribeInputButton(),
                    spacing=1,
                    align_y=0,
                ),
                margin_top=2,
                margin_x=1,
                spacing=2,
            )

        # Return the complete footer structure
        return rio.Column(
            # Divider line above the footer
            rio.Rectangle(
                fill=self.session.theme.neutral_color,
                min_height=0.1,  # Thin line for visual separation
            ),
            content,  # Main footer content
            margin_y=5,  # Add vertical margin around the footer
        )


# </component>



from __future__ import annotations

import dataclasses
import typing as t

import typing_extensions as te
from uniserde import JsonDoc

import rio

from .. import deprecations, utils
from .fundamental_component import FundamentalComponent

__all__ = ["FlowContainer"]


@t.final
class FlowContainer(FundamentalComponent):
    """
    A container whose children will be rearranged to fill the available space.

    `FlowContainer` is similar to `rio.Row` and `rio.Column`, except it
    automatically reflows its children if not enough space is available, similar
    to how text reflows in a word processor.

    ## Attributes

    `children`: The components to place inside the container.

    `row_spacing`: The vertical spacing between the children.

    `column_spacing`: The horizontal spacing between the children.

    `spacing`: Controls both the horizontal and the vertical spacing
        simultaneously.

    `justify`: The horizontal alignment of the children. Possible values are:
        `"left"`, `"center"`, `"right"`, `"justified"` and `"grow"`.


    ## Examples

    This minimal example will show two children horizontally next to one
    another, but reflow to vertical if the window is too small:

    ```python
    rio.FlowContainer(
        rio.Text("Hello"),
        rio.Text("World!"),
        column_spacing=1,
    )
    ```
    """

    children: list[rio.Component]
    _: dataclasses.KW_ONLY
    spacing: float | None
    row_spacing: float | None
    column_spacing: float | None
    justify: t.Literal["left", "center", "right", "justify", "grow"]

    def __init__(
        self,
        *children: rio.Component,
        spacing: float | None = None,
        row_spacing: float | None = None,
        column_spacing: float | None = None,
        justify: t.Literal[
            "left", "center", "right", "justify", "grow"
        ] = "left",
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        assert isinstance(children, tuple), children

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.children = list(children)
        self.spacing = spacing
        self.row_spacing = row_spacing
        self.column_spacing = column_spacing
        self.justify = justify

    def add(self, child: rio.Component) -> te.Self:
        """
        Appends a child component.

        Appends a child component to the end and then returns the
        `FlowContainer`, which makes method chaining possible:

        ```python
        rio.FlowContainer().add(child1).add(child2)
        ```


        ## Parameters

        `child`: The child component to append.
        """
        self.children.append(child)
        return self

    def _custom_serialize_(self) -> JsonDoc:
        result: JsonDoc = {
            "row_spacing": utils.first_non_none(
                self.row_spacing,
                self.spacing,
                0,
            ),
            "column_spacing": utils.first_non_none(
                self.column_spacing,
                self.spacing,
                0,
            ),
        }

        if self.justify == "justified":
            deprecations.warn(
                since="0.9.2",
                message=f'`justify="justified"` of `rio.FlowContainer` has been renamed. Please use `justify="justify"` instead.',
            )
            result["justify"] = "justify"

        return result


FlowContainer._unique_id_ = "FlowContainer-builtin"



import dataclasses
import typing as t

from uniserde import JsonDoc

from ..utils import URL
from .fundamental_component import FundamentalComponent

__all__ = ["Webview"]


@t.final
class Webview(FundamentalComponent):
    """
    Displays a website or renders HTML.

    `Webview` takes a URL or HTML markup as input and displays the website
    or the rendered HTML in your app.


    ## Attributes

    `content`: The URL of the website you want to display, or the HTML
        you want to render.

    `enable_pointer_events`: Whether the `Webview` component (and its contents)
        are clickable.

    `resize_to_fit_content`: Whether the `Webview` component should automatically
        update its size to match the size of its content. Note that this won't
        work if the displayed website's domain doesn't match your own domain.


    ## Examples

    This will display a website based on its URL:

    ```python
    rio.Webview(
        rio.URL("https://www.example.com"),
    )
    ```

    While this will render the given HTML markup:

    ```python
    rio.Webview("<html><body>Hello World</body></html>")
    ```

    The HTML doesn't necessarily have to be an entire website; something
    like this will also work just fine:

    ```python
    rio.Webview("<p>Hello World</p>")
    ```
    """

    content: URL | str
    _: dataclasses.KW_ONLY
    enable_pointer_events: bool = True
    resize_to_fit_content: bool = True

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "content": str(self.content),
        }


Webview._unique_id_ = "Webview-builtin"
